"""
Helper functions for viewing the block visually.

Each of the functions in visualization take a block and a file descriptor. The functions
provided write the block as a given visual format to the file.
"""

from __future__ import annotations

import collections
from typing import TYPE_CHECKING, Callable

from pyrtl.core import Block, LogicNet, working_block
from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError
from pyrtl.wire import Const, Input, Output, Register, WireVector

if TYPE_CHECKING:
    from pyrtl.simulation import SimulationTrace


def net_graph(block: Block = None, split_state: bool = False):
    """Return a graph representation of the given :class:`Block`.

    The graph has the following form::

        {
          node1: { nodeA: [edge1A_1, edge1A_2], nodeB: [edge1B]},
          node2: { nodeB: [edge2B],             nodeC: [edge2C_1, edge2C_2]},
          ...
        }

    aka: ``edges = graph[source][dest]``

    Each node can be either a :class:`LogicNet` or a :class:`WireVector` (e.g. an
    :class:`Input`, an :class:`Output`, a :class:`Const` or even an undriven
    :class:`WireVector` (which acts as a source or sink in the network). Each edge is a
    :class:`WireVector` or derived type (:class:`Input`, :class:`Output`,
    :class:`Register`, etc.). Note that :class:`Inputs<Input>`, :class:`Consts<Const>`,
    and :class:`Outputs<Output>` will be both "node" and "edge".
    :class:`WireVectors<WireVector>` that are not connected to any nets are not returned
    as part of the graph.

    .. note::

        Consider using :ref:`gate_graphs` instead.

    :param block: :class:`Block` to use (defaults to current :ref:`working_block`).
    :param split_state: If ``True``, split connections to/from a register update net;
        this means that registers will be appear as source nodes of the network, and
        ``r`` nets (i.e. the logic for setting :attr:`Register.next`) will be treated as
        sink nodes of the network.
    """
    # FIXME: make it not try to add unused wires (issue #204)
    block = working_block(block)

    # self.sanity_check()
    graph = {}

    # add all of the nodes
    for net in block.logic:
        graph[net] = {}

    wire_src_dict, wire_dst_dict = block.net_connections()
    dest_set = set(wire_src_dict.keys())
    arg_set = set(wire_dst_dict.keys())
    dangle_set = dest_set.symmetric_difference(arg_set)
    for w in dangle_set:
        graph[w] = {}
    if split_state:
        for w in block.wirevector_subset(Register):
            graph[w] = {}

    # add all of the edges
    for w in dest_set | arg_set:
        try:
            _from = wire_src_dict[w]
        except Exception:
            _from = w  # e.g. an Input/Const
        if split_state and isinstance(w, Register):
            _from = w

        try:
            _to_list = wire_dst_dict[w]
        except Exception:
            graph[_from][w] = [w]  # e.g. an Output
        else:
            for _to in _to_list:
                graph[_from][_to] = list(filter(lambda arg: arg is w, _to.args))

    return graph


# -----------------------------------------------------------------
#    ___  __   ___
#     |  / _` |___
#     |  \__> |


def _trivialgraph_default_namer(thing, is_edge=True):
    """Returns a "good" string for thing in printed graphs."""
    if is_edge:
        if thing.name is None or thing.name.startswith("tmp"):
            return ""
        return "/".join([thing.name, str(len(thing))])
    if isinstance(thing, Const):
        return str(thing.val)
    if isinstance(thing, WireVector):
        return thing.name or "??"
    try:
        return thing.op + str(thing.op_param or "")
    except AttributeError as exc:
        msg = f'no naming rule for "{thing}"'
        raise PyrtlError(msg) from exc


def output_to_trivialgraph(
    file,
    namer: Callable[[WireVector | LogicNet, bool], str] = _trivialgraph_default_namer,
    block: Block = None,
    split_state: bool = False,
):
    """Walk the block and output it in `trivial graph format
    <https://en.wikipedia.org/wiki/Trivial_Graph_Format>`_ to the open file.

    :param file: Open file to write to.
    :param namer: A function that takes in an object (a :class:`WireVector` or
        :class:`LogicNet`) as the first argument and a boolean ``is_edge`` as the second
        that is set ``True`` if the object is a :class:`WireVector`, and returns a
        string representing that object.
    :param block: :class:`Block` to use (defaults to current :ref:`working_block`).
    :param split_state: If ``True``, split connections to/from a register update net;
        this means that registers will be appear as source nodes of the network, and
        ``r`` :class:`LogicNets<LogicNet>` (i.e. the logic for setting
        :attr:`Register.next`) will be treated as sink nodes of the network.
    """
    graph = net_graph(block, split_state)
    node_index_map = {}  # map node -> index

    # print the list of nodes
    for index, node in enumerate(graph):
        print(index, namer(node, is_edge=False), file=file)
        node_index_map[node] = index

    print("#", file=file)

    # print the list of edges
    for _from in graph:
        for _to in graph[_from]:
            from_index = node_index_map[_from]
            to_index = node_index_map[_to]
            for edge in graph[_from][_to]:
                print(from_index, to_index, namer(edge), file=file)


# -----------------------------------------------------------------
#     __   __        __              __
#    / _` |__)  /\  |__) |__| \  / |  /
#    \__> |  \ /~~\ |    |  |  \/  | /__


def _default_edge_namer(
    edge: WireVector,
    is_to_splitmerge: bool = False,
    extra_edge_info: dict[WireVector, str] | None = None,
):
    """
    A function for naming an edge for use in the ``graphviz`` graph.

    :param edge: the edge (i.e. :class:`WireVector` or deriving class)
    :param is_to_splitmerge: if the node to which the edge points
        is a ``select`` or ``concat`` operation
    :param extra_edge_info: a map from edge to any additional data you want
        to print associated with it (e.g. timing data)
    :return: a function that can be called by graph namer function you pass
        in to ``block_to_graphviz_string``
    """

    name = "" if edge.name is None else "/".join([edge.name, str(len(edge))])
    if extra_edge_info and edge in extra_edge_info:
        # Always label an edge if present in the extra_edge_info map
        name = name + " (" + str(extra_edge_info[edge]) + ")"
    elif (
        edge.name is None
        or edge.name.startswith("tmp")
        or isinstance(edge, (Input, Output, Const, Register))
    ):
        name = ""

    penwidth = 2 if len(edge) == 1 else 6
    arrowhead = "none" if is_to_splitmerge else "normal"
    return f'[label="{name}", penwidth="{penwidth}", arrowhead="{arrowhead}"]'


def _default_node_namer(
    node: WireVector,
    split_state: bool = False,
    extra_node_info: dict[WireVector, str] | None = None,
):
    """
    A function for naming a node for use in the ``graphviz`` graph.

    :param node: the node (i.e. :class:`WireVector` or deriving class, or a logic net)
    :param split_state: if ``True``, split connections to/from a register update net;
        this means that registers will be appear as source nodes of the network, and 'r'
        nets (i.e. the logic for setting a register's next value) will be treated as
        sink nodes of the network.
    :param extra_node_info: a map from node to any additional data you want to print
        associated with it (e.g. delay data)

    :return: a function that can be called by graph namer function you pass in to
             :func:`block_to_graphviz_string`
    """

    def label(v):
        if extra_node_info and node in extra_node_info:
            v = v + " (" + str(extra_node_info[node]) + ")"
        return v

    if isinstance(node, Const):
        name = node.name + ": " if not node.name.startswith("const_") else ""
        return (
            f'[label="{label(name + str(node.val))}", shape=circle, '
            "fillcolor=lightgrey]"
        )
    if isinstance(node, Input):
        return f'[label="{label(node.name)}", shape=invhouse, fillcolor=coral]'
    if isinstance(node, Output):
        return f'[label="{label(node.name)}", shape=house, fillcolor=lawngreen]'
    if isinstance(node, Register):
        return f'[label="{label(node.name)}", shape=square, fillcolor=gold]'
    if isinstance(node, WireVector):
        return f'[label="{label(node.name)}", shape=circle, fillcolor=none]'
    try:
        if node.op == "&":
            return '[label="{}"]'.format(label("and"))
        if node.op == "|":
            return '[label="{}"]'.format(label("or"))
        if node.op == "^":
            return '[label="{}"]'.format(label("xor"))
        if node.op == "~":
            return '[label="{}", shape=invtriangle]'.format(label("not"))
        if node.op == "x":
            return '[label="{}", shape=invtrapezium]'.format(label("mux"))
        if node.op == "s":
            # node.op_param is a tuple of the selected bits to pull from the argument
            # wire, so it could look something like (0,0,0,0,0,0,0), meaning dest wire
            # is going to be a concatenation of the zero-th bit of the argument wire, 7
            # times.
            selLower = node.op_param[0]
            selUpper = node.op_param[-1]
            if len(node.op_param) == 1:
                bits = f"[{selLower}]"
            elif node.op_param == tuple(range(selLower, selUpper + 1)):  # consecutive
                bits = f"[{selUpper}:{selLower}]"
            elif all(
                ix == node.op_param[0] for ix in node.op_param[1:]
            ):  # all the same
                bits = f"[{node.op_param[0]}]*{len(node.op_param)}"
            else:
                bits = "bits" + str(tuple(reversed(node.op_param)))
            return f'[label="{label(bits)}", fillcolor=azure1, height=.25, width=.25]'
        if node.op in "c":
            return '[label="{}", height=.1, width=.1]'.format(label("concat"))
        if node.op == "r":
            name = node.dests[0].name or ""
            name = (f"{name}.next") if split_state else name
            return f'[label="{label(name)}", shape=square, fillcolor=gold]'
        if node.op == "w":
            return '[label="{}", height=.1, width=.1]'.format(label(""))
        if node.op in "m@":
            name = node.op_param[1].name
            if name.startswith("tmp"):
                name = ""
            else:
                name = "(" + name + ")"
            return f'[label="{label(node.op + name)}"]'
        return '[label="{}"]'.format(label(node.op + str(node.op_param or "")))
    except AttributeError as exc:
        msg = f'no naming rule for "{node}"'
        raise PyrtlError(msg) from exc


def _graphviz_default_namer(
    thing: WireVector | LogicNet,
    is_edge: bool,
    is_to_splitmerge: bool,
    split_state: bool,
    node_namer=_default_node_namer,
    edge_namer=_default_edge_namer,
):
    """Returns a "good" Graphviz label for thing.

    :param thing: The edge (:class:`WireVector`) or node (:class:`LogicNet` or
        :class:`Input`/:class:`Output`/:class:`Const`) to name
    :param is_edge: ``True`` if thing is an edge
    :param is_to_splitmerge: if the node to which the edge points is a ``select`` or
        concat operation
    :param split_state: If ``True``, visually split the connections to/from a register
        update net.
    :param node_namer: A function mapping a node to a label; one of its arguments is a
        dict mapping nodes to nodes to additional user-supplied information.
    :param edge_namer: A function mapping an edge to a label; one of its arguments is a
        dict mapping nodes to nodes to additional user-supplied information.

    :return: A function that knows how to label each element in the graph, which can be
             passed to :func:`output_to_graphviz` or :func:`block_to_graphviz_string`
    """
    if is_edge:
        return edge_namer(thing, is_to_splitmerge=is_to_splitmerge)
    return node_namer(thing, split_state=split_state)


def graphviz_detailed_namer(
    extra_node_info: dict | None = None, extra_edge_info: dict | None = None
):
    """Returns a detailed Graphviz namer that prints extra information about nodes/edges
    in the given maps.

    If both :class:`dict` arguments are ``None``, the returned namer behaves identically
    to the default Graphviz namer.

    :param extra_node_info: A :class:`dict` from node to additional data about that
        node. The additional data will be converted to :class:`str` and printed next to
        the node's label.
    :param extra_edge_info: A :class:`dict` from edge to additional data about that
        edge. The additional data will be converted to :class:`str` and printed next to
        the edge's label.

    :return: A function to label each element in the graph, which can be used as
             :func:`output_to_graphviz` or :func:`block_to_graphviz_string`'s ``namer``.
    """

    def node_namer(node, split_state):
        return _default_node_namer(node, split_state, extra_node_info)

    def edge_namer(edge, is_to_splitmerge):
        return _default_edge_namer(edge, is_to_splitmerge, extra_edge_info)

    def namer(thing, is_edge, is_to_splitmerge, split_state):
        return _graphviz_default_namer(
            thing,
            is_edge,
            is_to_splitmerge,
            split_state,
            node_namer=node_namer,
            edge_namer=edge_namer,
        )

    return namer


def output_to_graphviz(
    file,
    block: Block = None,
    namer=_graphviz_default_namer,
    split_state: bool = True,
    maintain_arg_order: bool = False,
):
    """Walk the :class:`Block` and output it in `Graphviz <https://graphviz.org/>`_
    format to the open file.

    ``output_to_graphviz`` writes a file containing a directed graph in the format
    expected by `Graphviz <https://graphviz.org/>`_, specifically in the :command:`dot`
    format. Once Graphviz is installed, the resulting graph file can be rendered to a
    ``.pdf`` file with::

        dot -Tpdf output.dot > output.pdf

    :param file: Open file to write to.
    :param block: :class:`Block` to use (defaults to current :ref:`working_block`)
    :param namer: Function used to label each edge and node; see
        :func:`block_to_graphviz_string` for more information.
    :param split_state: If ``True``, visually split the connections to/from a
        :class:`Register` update net.
    :param maintain_arg_order: If ``True``, add ordering constraints so incoming edges
        are ordered left-to-right for nets where argument order matters (e.g. ``<``).
        Keeping this as ``False`` results in a cleaner, though less visually precise,
        graphical output.
    """
    print(
        block_to_graphviz_string(block, namer, split_state, maintain_arg_order),
        file=file,
    )


def block_to_graphviz_string(
    block: Block = None,
    namer=_graphviz_default_namer,
    split_state: bool = True,
    maintain_arg_order: bool = False,
):
    """Return a Graphviz string for the ``block``.

    The normal namer function will label user-named wires with their names and label the
    nodes (:class:`LogicNets<LogicNet>` or :class:`Input`/:class:`Output`/:class:`Const`
    terminals) with their operator symbol or name/value, respectively. If custom
    information about each node in the graph is desired, you can pass in a custom namer
    function which must have the same signature as the default namer,
    :func:`_graphviz_default_namer`.

    However, we recommend you instead pass in a call to :func:`graphviz_detailed_namer`,
    supplying it with your own :class:`dicts<dict>` mapping wires and nodes to labels.
    For any wire/node found in these maps, that additional information will be printed
    in parentheses alongside the node in the ``graphviz`` graph.

    For example, if you wanted to print the delay of each wire and the fanout of each
    gate, you could pass in two maps to the :func:`graphviz_detailed_namer` call, which
    returns a namer function that can subsequently be passed to
    :func:`output_to_graphviz` or :func:`block_to_graphviz_string`::

        node_fanout = {n: f"Fanout: {my_fanout_func(n)}"
                       for n in working_block().logic}
        wire_delay = {w: f"Delay: {my_delay_func(w):.2f}"
                      for w in working_block().wirevector_set}

        with open("out.gv", "w") as f:
            output_to_graphviz(
                f, namer=graphviz_detailed_namer(node_fanout, wire_delay))

    :param namer: A function mapping graph objects (wires/logic nets) to labels. If you
        want a more detailed namer, pass in a call to :func:`graphviz_detailed_namer`.
    :param block: :class:`Block` to use (defaults to current :ref:`working_block`)
    :param bool split_state: If ``True``, split connections to/from a :class:`Register`
        update net; this means that registers will be appear as source nodes of the
        network, and ``r`` nets (i.e. the logic for setting :attr:`Register.next`) will
        be treated as sink nodes of the network.
    :param bool maintain_arg_order: If ``True``, will add ordering constraints so
        incoming edges are ordered left-to-right for nets where argument order matters
        (e.g. ``<``). Keeping this as ``False`` results in a cleaner, though less
        visually precise, graphical output.
    """
    graph = net_graph(block, split_state)
    node_index_map = {}  # map node -> index

    rstring = """\
digraph g {
    graph [splines="spline", outputorder="edgesfirst"];
    node [shape=circle, style=filled, fillcolor=lightblue1,
        fontcolor=black, fontname=helvetica, penwidth=0,
        fixedsize=shape];
    edge [labelfloat=false, penwidth=2, color=deepskyblue, arrowsize=.5];
"""
    from pyrtl.importexport import _natural_sort_key

    def _node_sort_key(node):
        # If a LogicNet and a wire share the same name, we want the LogicNet to sort
        # first, so we arbitrarily 'A' and 'B' suffixes to break ties.
        if isinstance(node, LogicNet):
            if node.op == "@":
                key = str(node.args[2]) + "A"
            else:
                key = node.dests[0].name + "A"
        else:
            key = node.name + "B"
        return _natural_sort_key(key)

    # print the list of nodes
    for index, node in enumerate(sorted(graph.keys(), key=_node_sort_key)):
        label = namer(node, False, False, split_state)
        rstring += f"    n{index} {label};\n"
        node_index_map[node] = index

    # print the list of edges
    srcs = collections.defaultdict(list)
    for _from in sorted(graph.keys(), key=_node_sort_key):
        for _to in sorted(graph[_from].keys(), key=_node_sort_key):
            from_index = node_index_map[_from]
            to_index = node_index_map[_to]
            for edge in graph[_from][_to]:
                is_to_splitmerge = hasattr(_to, "op") and _to.op in "cs"
                label = namer(edge, True, is_to_splitmerge, False)
                rstring += f"    n{from_index} -> n{to_index} {label};\n"
                srcs[_to].append((_from, edge))

    # Maintain left-to-right order of incoming wires for nets where order matters. This
    # won't be visually perfect sometimes (especially for a wire used twice in a net's
    # argument list), but for the majority of cases this will improve the visualization.
    def index_of(w, args):
        # Special helper so we compare id rather than using builtin operators
        for i, arg in enumerate(args):
            if w is arg:
                return i
        msg = "Expected to find wire in set of args"
        raise PyrtlInternalError(msg)

    if maintain_arg_order:
        block = working_block(block)
        for net in sorted(block.logic_subset(op="c-<>x@"), key=_node_sort_key):
            args = [(node_index_map[n], wire) for (n, wire) in srcs[net]]
            args.sort(key=lambda t: index_of(t[1], net.args))
            s = " -> ".join([f"n{n}" for n, _ in args])
            rstring += "    {\n"
            rstring += "        rank=same;\n"
            rstring += "        edge[style=invis];\n"
            rstring += "        " + s + ";\n"
            rstring += "        rankdir=LR;\n"
            rstring += "    }\n"

    rstring += "}\n"
    return rstring


# -----------------------------------------------------------------
#     __        __
#    /__` \  / / _`
#    .__/  \/  \__>


def output_to_svg(file, block: Block = None, split_state: bool = True):
    """Output the block as an SVG to the open file.

    :param file: Open file to write to.
    :param block: :class:`Block` to use (defaults to current :ref:`working_block`).
    :param split_state: If ``True``, visually split the connections to/from a register
        update net.
    """
    print(block_to_svg(block, split_state), file=file)


def block_to_svg(
    block: Block = None, split_state: bool = True, maintain_arg_order: bool = False
):
    """Return an SVG for the block.

    :param block: :class:`Block` to use (defaults to current :ref:`working_block`).
    :param split_state: If ``True``, visually split the connections to/from a register
        update net.
    :param maintain_arg_order: If ``True``, will add ordering constraints so incoming
        edges are ordered left-to-right for nets where argument order matters (e.g.
        ``<``). Keeping this as ``False`` results in a cleaner, though less visually
        precise, graphical output.

    :return: The SVG representation of the :class:`Block`.
    """
    try:
        from graphviz import Source

        src = Source(
            block_to_graphviz_string(
                block, split_state=split_state, maintain_arg_order=maintain_arg_order
            )
        )
        try:
            svg = src._repr_image_svg_xml()
        except AttributeError:
            # py-graphviz 0.18.3 or earlier
            return src._repr_svg_()
        else:
            # py-graphviz 0.19 or later
            return svg
    except ImportError as exc:
        msg = 'need graphviz installed (try "pip install graphviz")'
        raise PyrtlError(msg) from exc


# -----------------------------------------------------------------
#         ___
#    |__|  |  |\/| |
#    |  |  |  |  | |___


def trace_to_html(
    simtrace: SimulationTrace,
    trace_list: list[str] | None = None,
    sortkey=None,
    repr_func: Callable[[int], str] = hex,
    repr_per_name: dict[str, Callable[[int], str]] | None = None,
) -> str:
    """Return a HTML block showing the trace.

    :param simtrace: A trace to render in HTML.
    :param trace_list: (optional) A list of wires to display.
    :param sortkey: (optional) The key with which to sort the ``trace_list``.
    :param repr_func: Function to use for representing each value in the trace. Examples
        include :func:`hex`, :func:`oct`, :func:`bin`, and :class:`str` (for decimal),
        :func:`val_to_signed_integer` (for signed decimal) or the function returned by
        :func:`enum_name` (for :class:`~enum.IntEnum`). Defaults to :func:`hex`.
    :param repr_per_name: Map from signal name to a function that takes in the signal's
        value and returns a user-defined representation. If a signal name is not found
        in the map, the argument ``repr_func`` will be used instead.

    :return: An HTML block showing the trace.
    """

    from pyrtl.simulation import SimulationTrace, _trace_sort_key

    if repr_per_name is None:
        repr_per_name = {}
    if not isinstance(simtrace, SimulationTrace):
        msg = "first arguement must be of type SimulationTrace"
        raise PyrtlError(msg)

    trace = simtrace.trace
    if sortkey is None:
        sortkey = _trace_sort_key

    if trace_list is None:
        trace_list = sorted(trace, key=sortkey)

    wave_template = """\
<script type="WaveDrom">
{
  signal : [
%s
  ],
  config: { hscale: %d }
}
</script>
"""

    vallens = []  # For determining longest value length

    def extract(w):
        wavelist = []
        datalist = []
        last = None

        for value in trace[w]:
            if last == value:
                wavelist.append(".")
            else:
                f = repr_per_name.get(w)
                if f is not None:
                    wavelist.append("=")
                    datalist.append(str(f(value)))
                elif len(simtrace._wires[w]) == 1:
                    # int() to convert True/False to 0/1
                    wavelist.append(str(int(value)))
                else:
                    wavelist.append("=")
                    datalist.append(str(repr_func(value)))

                last = value

        wavestring = "".join(wavelist)
        datastring = ", ".join([f'"{data}"' for data in datalist])
        if repr_per_name.get(w) is None and len(simtrace._wires[w]) == 1:
            vallens.append(1)  # all are the same length
            return bool_signal_template % (w, wavestring)
        vallens.extend([len(data) for data in datalist])
        return int_signal_template % (w, wavestring, datastring)

    bool_signal_template = '    { name: "%s",  wave: "%s" },'
    int_signal_template = '    { name: "%s",  wave: "%s", data: [%s] },'
    signals = [extract(w) for w in trace_list]
    all_signals = "\n".join(signals)
    maxvallen = max(vallens)
    scale = (maxvallen // 5) + 1
    return wave_template % (all_signals, scale)
    # print(wave)



# # Example 1.2: `wire_struct` and `wire_matrix()`.
#
# This example demonstrates named slicing with `wire_struct` and `wire_matrix()`:
# * `wire_struct`: Named `WireVector` slices. Names are alphanumeric, like a Python
#   namedtuple.
# * `wire_matrix()`: Indexed `WireVector` slices. Indices are integers, like a Python
#   list.
import inspect

import pyrtl

# ## Motivation
#
# `wire_struct` and `wire_matrix()` are syntactic sugar. They make PyRTL code easier to
# read and write, but they do not introduce any new functionality. We start with a
# motivating example that does not use `wire_struct` or `wire_matrix()`, point out some
# common problems, then rewrite the example with `wire_struct` to address those
# problems.
#
# This example deals with 24-bit pixel data. These 24 bits are composed of three
# concatenated components:
#
#     pixel:
#      ┌───────┬───────┬───────┐
#      │  red  │ green │ blue  │
#      └┬─────┬┴┬─────┬┴┬─────┬┘
#      23    16 15    8 7     0
#
# In other words:
# * The `red` component is in `pixel[16:24]`
# * The `green` component is in `pixel[8:16]`
# * And the `blue` component is in `pixel[0:8]`
#
# Accept a 24-bit pixel as a PyRTL `Input`:
pixel_in = pyrtl.Input(name="pixel_in", bitwidth=24)

# Split the pixel data into `red`, `green`, and `blue` components. The indices are
# manually calculated and hardcoded below. These calculations are prone to off-by-one
# errors. If we wanted to change the pixel data format (for example, adding an 8-bit
# `alpha` channel), we may need to update all these indices.
red_in = pixel_in[16:24]
red_in.name = "red_in"
green_in = pixel_in[8:16]
green_in.name = "green_in"
blue_in = pixel_in[0:8]
blue_in.name = "blue_in"

# Do some calculations on the individual components:
red_out = red_in + 0x11
green_out = green_in + 0x22
blue_out = blue_in + 0x33

# `red_out`, `green_out`, and `blue_out` are 9 bits wide, so they must be truncated back
# down to 8 bits before concatenating.
assert red_out.bitwidth == 9
assert green_out.bitwidth == 9
assert blue_out.bitwidth == 9

# Now `concat()` the processed components into a new output pixel. The components must
# be specified in the correct order, and we must `truncate()` them to the correct
# bitwidth.
pixel_out = pyrtl.concat(
    red_out.truncate(8), green_out.truncate(8), blue_out.truncate(8)
)
assert pixel_out.bitwidth == 24
pixel_out.name = "pixel_out"

# Test this slicing and concatenation logic.
print("Pixel slicing and concatenation without wire_struct:")
sim = pyrtl.Simulation()
sim.step_multiple(provided_inputs={"pixel_in": [0x112233, 0xAABBCC]})
sim.tracer.render_trace(
    trace_list=["pixel_in", "red_in", "green_in", "blue_in", "pixel_out"]
)

# ## Introducing `wire_struct`
#
# Start over, and rebuild the example with `wire_struct`.
pyrtl.reset_working_block()


# This `Pixel` class is the single source of truth that specifies how the `red`,
# `green`, and `blue` components are packed into 24 bits. A `Pixel` instance can do one
# of two things:
#
# 1. Slice a 24-bit value into `red`, `green`, and `blue` components, OR
# 2. Pack `red`, `green`, and `blue` components into a 24-bit value.
#
# Because this definition is the single source of truth, we can change the pixel data
# format just by changing this class definition. For example, try making the `red`
# component the last component, instead of the first component.
@pyrtl.wire_struct
class Pixel:
    # The most significant bits are specified first, so the `red` component is the 8
    # most significant bits, indices 16..23
    red: 8
    # The `green` component is the 8 middle bits, indices 8..15
    green: 8
    # The `blue` component is the 8 least significant bits, indices 0..7
    blue: 8


# `pixel_in` is a `Pixel` constructed from a 24-bit `Input`, option (1) above. The
# 24-bit `Input` will be sliced into 8-bit components, and those components can be
# manipulated as `pixel_in.red`, `pixel_in.green`, and `pixel_in.blue` in the Python
# code below.
#
# Because this `Pixel` has a name, the 8-bit components will be assigned the names
# `"pixel_in.red"`, `"pixel_in.green"`, and `"pixel_in.blue"`. These component names are
# included in the `trace_list` below.
pixel_in = Pixel(name="pixel_in", concatenated_type=pyrtl.Input)
assert pixel_in.bitwidth == 24
assert pixel_in.red.bitwidth == 8
assert pixel_in.green.bitwidth == 8
assert pixel_in.blue.bitwidth == 8

# Repeat our calculations with the named components. Components are accessed with "dot"
# notation (`.green`), like a Python namedtuple. There are no more hardcoded indices
# like `[8:16]` in this example.
red_out = pixel_in.red + 0x11
green_out = pixel_in.green + 0x22
blue_out = pixel_in.blue + 0x33

# `pixel_out` is a `Pixel` constructed from components, option (2) above. Components
# with more than 8 bits are implicitly truncated to 8 bits.
#
# Python keyword arguments are used for `red`, `green`, and `blue` here, so the order
# does not matter. We arbitrarily provide the value of the `green` component first.
pixel_out = Pixel(name="pixel_out", green=green_out, red=red_out, blue=blue_out)
assert pixel_out.bitwidth == 24

# Test this new version of our slicing and concatenation logic.
print("\nPixel slicing and concatenation with wire_struct:")
sim = pyrtl.Simulation()
sim.step_multiple(provided_inputs={"pixel_in": [0x112233, 0xAABBCC]})
sim.tracer.render_trace(
    trace_list=[
        "pixel_in",
        "pixel_in.red",
        "pixel_in.green",
        "pixel_in.blue",
        "pixel_out",
    ]
)

# ## Introducing `wire_matrix()`
#
# `wire_matrix()` is like `wire_struct`, except the components are indexed by integers,
# like a list.
#
# Start over.
pyrtl.reset_working_block()

# `PixelPair` is a pair of two `Pixels`. This also shows how `wire_matrix()` and
# `wire_struct` work together - `PixelPair` is a `wire_struct` nested in a
# `wire_matrix()`.
PixelPair = pyrtl.wire_matrix(component_schema=Pixel, size=2)

# `wire_matrix()` returns a class!
assert inspect.isclass(PixelPair)

# Create a `PixelPair` that decomposes a 48-bit input into two 24-bit `Pixels`. Each
# 24-bit `Pixel` can be further decomposed into 8-bit `red`, `green`, and `blue`
# components. This `PixelPair` has been assigned a name, so all its components are
# named, as seen in the simulation trace below.
pixel_pair_in = PixelPair(name="pixel_pair_in", concatenated_type=pyrtl.Input)
assert pixel_pair_in.bitwidth == 48
assert pixel_pair_in[0].bitwidth == 24
assert pixel_pair_in[0].red.bitwidth == 8

# Do some calculations on the `Pixel` components. Components are accessed with
# square-bracket notation (`[1]`), like a Python list.
pixel_pair_out_values = [
    Pixel(
        red=pixel_pair_in[0].red + 0x10,
        green=pixel_pair_in[0].green,
        blue=pixel_pair_in[0].blue,
    ),
    pixel_pair_in[1] + 1,
]

# Pack the new component values into a new `PixelPair`.
pixel_pair_out = PixelPair(name="pixel_pair_out", values=pixel_pair_out_values)

# Test out `PixelPair`.
print("\nPixelPair slicing and concatenating with wire_matrix:")
sim = pyrtl.Simulation()
sim.step_multiple(provided_inputs={"pixel_pair_in": [0x112233AABBCC, 0x445566778899]})
sim.tracer.render_trace(
    trace_list=[
        "pixel_pair_in",
        "pixel_pair_in[0]",
        "pixel_pair_in[0].red",
        "pixel_pair_in[0].green",
        "pixel_pair_in[0].blue",
        "pixel_pair_in[1]",
        "pixel_pair_in[1].red",
        "pixel_pair_in[1].green",
        "pixel_pair_in[1].blue",
        "pixel_pair_out",
    ]
)

# These examples hopefully show how `wire_matrix()` and `wire_struct` result in code
# that's easier to write correctly and easier to read.
#
# `wire_struct` and `wire_matrix()` have many more features, but this example should
# demonstrate the main ideas. See the documentation for more details:
#
# https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.wire_struct
#
# https://pyrtl.readthedocs.io/en/latest/helpers.html#pyrtl.wire_matrix



"""
Defines PyRTL memories. These blocks of memories can be read (potentially async) and
written (sync)

MemBlocks supports any number of the following operations:

- read: `d = mem[address]`

- write: `mem[address] <<= d`

- write with an enable: `mem[address] <<= MemBlock.EnabledWrite(d, enable=we)`

Based on the number of reads and writes a memory will be inferred with the correct
number of ports to support that
"""

from __future__ import annotations

import collections
import numbers
import types
from typing import NamedTuple

from pyrtl.core import Block, LogicNet, _NameIndexer, working_block
from pyrtl.corecircuits import as_wires
from pyrtl.helperfuncs import infer_val_and_bitwidth
from pyrtl.pyrtlexceptions import PyrtlError
from pyrtl.wire import Const, WireVector, WireVectorLike, next_tempvar_name

# ------------------------------------------------------------------------
#
#         ___        __   __          __        __   __
#   |\/| |__   |\/| /  \ |__) \ /    |__) |    /  \ /  ` |__/
#   |  | |___  |  | \__/ |  \  |     |__) |___ \__/ \__, |  \
#


_memIndex = _NameIndexer()

_MemAssignment = collections.namedtuple("_MemAssignment", "rhs, is_conditional")
"""_MemAssignment is the type returned from assignment by |= or <<="""


def _reset_memory_indexer():
    global _memIndex
    _memIndex = _NameIndexer()


class _MemIndexed(WireVector):
    """Object used internally to route memory assigns correctly.

    The normal PyRTL user should never need to be aware that this class exists, hence
    the underscore in the name. It presents a very similar interface to WireVectors (all
    of the normal wirevector operations should still work), but if you try to *set* the
    value with <<= or |= then it will generate a _MemAssignment object rather than the
    normal wire assignment.
    """

    def __init__(self, mem, index):
        self.mem = mem
        self.index = index
        self.wire = None

    def __ilshift__(self, other):
        return _MemAssignment(rhs=other, is_conditional=False)

    def __ior__(self, other):
        return _MemAssignment(rhs=other, is_conditional=True)

    def _two_var_op(self, other, op):
        return as_wires(self)._two_var_op(other, op)

    def __invert__(self):
        return as_wires(self).__invert__()

    def __getitem__(self, item):
        return as_wires(self).__getitem__(item)

    def __len__(self):
        return self.mem.bitwidth

    def sign_extended(self, bitwidth):
        return as_wires(self).sign_extended(bitwidth)

    def zero_extended(self, bitwidth):
        return as_wires(self).zero_extended(bitwidth)

    @property
    def name(self):
        return as_wires(self).name

    @name.setter
    def name(self, n):
        as_wires(self).name = n


class MemBlock:
    """``MemBlock`` is the object for specifying block memories.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    ``MemBlock`` can be indexed like an array for reads and writes. Example::

        >>> mem = pyrtl.MemBlock(bitwidth=8, addrwidth=2)

        >>> # Write to each address, starting from address 1.
        >>> write_addr = pyrtl.Register(name="write_addr", bitwidth=2, reset_value=1)
        >>> write_addr.next <<= write_addr + 1

        >>> # Read from each address, starting from address 0.
        >>> read_addr = pyrtl.Register(name="read_addr", bitwidth=2)
        >>> read_addr.next <<= read_addr + 1

        >>> read_data = pyrtl.Output(name="read_data")
        >>> read_data <<= mem[read_addr]  # Creates a read port.
        >>> mem[write_addr] <<= write_addr + 10  # Creates a write port.

        >>> sim = pyrtl.Simulation()
        >>> sim.step_multiple(nsteps=6)
        >>> sim.tracer.trace["write_addr"]
        [1, 2, 3, 0, 1, 2]
        >>> sim.tracer.trace["read_addr"]
        [0, 1, 2, 3, 0, 1]
        >>> sim.tracer.trace["read_data"]
        [0, 11, 12, 13, 10, 11]

    .. doctest only::

        >>> pyrtl.reset_working_block()

    When the address of a memory is assigned to using an :class:`EnabledWrite` object,
    data will only be written to the memory when the ``EnabledWrite``'s
    :attr:`~EnabledWrite.enable` ``WireVector`` is set to high (``1``). In the following
    example, the ``MemBlock`` is only written when ``write_addr`` is odd::

        >>> mem = pyrtl.MemBlock(bitwidth=8, addrwidth=2)

        >>> write_addr = pyrtl.Register(name="write_addr", bitwidth=2)
        >>> write_addr.next <<= write_addr + 1
        >>> mem[write_addr] <<= pyrtl.MemBlock.EnabledWrite(
        ...     enable=write_addr[0], data=write_addr + 10)

        >>> sim = pyrtl.Simulation()
        >>> sim.step_multiple(nsteps=6)
        >>> sorted(sim.inspect_mem(mem).items())
        [(1, 11), (3, 13)]

    Writes under :ref:`conditional_assignment` are automatically converted to
    :class:`EnabledWrites<EnabledWrite>`.

    .. _asynchronous_memories:

    Asynchronous Memories
    ---------------------

    It is best practice to have memory operations start on a rising clock edge if you
    want them to synthesize into efficient hardware, so ``MemBlocks`` are `synchronous`
    by default (``asynchronous=False``). ``MemBlocks`` will enforce this by checking
    that all their inputs are ready at each rising clock edge. This implies that all
    ``MemBlock`` inputs - the address to read/write, the data to write, and the
    write-enable bit - must be registers, inputs, or constants, unless you explicitly
    declare the memory as `asynchronous` with ``asynchronous=True``.

    Asynchronous memories can be convenient and tempting, but they are rarely a good
    idea. They can't be mapped to block RAMs in FPGAs and will be converted to registers
    by most design tools. They are not a realistic option for memories with more than a
    few hundred elements.

    Read and Write Ports
    --------------------

    Each read or write to the memory will create a new `port` (either a read port or
    write port respectively). By default memories are limited to 2 read ports and 1
    write port, to keep designs efficient by default, but those values can be changed
    with ``max_read_ports`` and ``max_write_ports``. Note that memories with many ports
    may not map to physical memories such as block RAMs or existing memory hardware
    macros.

    Default Values
    --------------

    In PyRTL :class:`Simulation`, all ``MemBlocks`` are zero-initialized by default.
    Initial data can be specified for each MemBlock in :meth:`Simulation.__init__`'s
    ``memory_value_map``.

    Simultaneous Read and Write
    ---------------------------

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    In PyRTL :class:`Simulation`, if the same address is read and written in the same
    cycle, the read will return the `last` value stored in the ``MemBlock``, not the
    newly written value. Example::

        >>> mem = pyrtl.MemBlock(addrwidth=1, bitwidth=1)
        >>> mem[0] <<= 1
        >>> read_data = pyrtl.Output(name="read_data", bitwidth=1)
        >>> read_data <<= mem[0]

        >>> # In the first cycle, read_data will be the default MemBlock data value
        >>> # (0), not the newly written value (1).
        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> sim.inspect("read_data")
        0

        # In the second cycle, read_data will be the newly written value (1).
        >>> sim.step()
        >>> sim.inspect("read_data")
        1

    Mapping ``MemBlocks`` to Hardware
    ---------------------------------

    Synchronous ``MemBlocks`` can generally be mapped to FPGA block RAMs and similar
    hardware, but there are many pitfalls:

    #. ``asynchronous=False`` is generally necessary, but may not be sufficient, for
       mapping a design to FPGA block RAMs. Block RAMs may have additional timing
       constraints, like requiring register outputs for each block RAM.
       ``asynchronous=False`` only requires register inputs.

    #. Block RAMs may offer more or less read and write ports than ``MemBlock``'s
       defaults.

    #. Block RAMs may not zero-initialize by default.

    #. Block RAMs may implement simultaneous reads and writes in different ways.
    """

    # FIXME: write ports assume that only one port is under control of the conditional
    class EnabledWrite(NamedTuple):
        """Generates logic to conditionally enable a write port."""

        data: WireVector
        """Data to write."""
        enable: WireVector
        """Single-bit ``WireVector`` indicating if a write should occur."""

    def __init__(
        self,
        bitwidth: int,
        addrwidth: int,
        name: str = "",
        max_read_ports: int = 2,
        max_write_ports: int = 1,
        asynchronous: bool = False,
        block: Block = None,
    ):
        """Create a PyRTL read-write memory.

        :param bitwidth: The bitwidth of each element in the memory.
        :param addrwidth: The number of bits used to address an element in the memory.
            The memory can store ``2 ** addrwidth`` elements.
        :param name: Name of the memory. Defaults to an autogenerated name.
        :param max_read_ports: limits the number of read ports each block can create;
            passing ``None`` indicates there is no limit.
        :param max_write_ports: limits the number of write ports each block can create;
            passing ``None`` indicates there is no limit.
        :param asynchronous: If ``False``, ensure that all memory inputs are registers,
            inputs, or constants. See :ref:`asynchronous_memories`.
        :param block: The block to add the MemBlock to, defaults to the
            :ref:`working_block`.
        """
        self.max_read_ports = max_read_ports
        self.num_read_ports = 0
        self.block = working_block(block)
        name = next_tempvar_name(name)

        if bitwidth <= 0:
            msg = "bitwidth must be >= 1"
            raise PyrtlError(msg)
        if addrwidth <= 0:
            msg = "addrwidth must be >= 1"
            raise PyrtlError(msg)

        self.bitwidth = bitwidth
        self.name = name
        self.addrwidth = addrwidth
        self.readport_nets = []
        self.id = _memIndex.next_index()
        self.asynchronous = asynchronous
        self.block._add_memblock(self)

        self.max_write_ports = max_write_ports
        self.num_write_ports = 0
        self.writeport_nets = []

    @property
    def read_ports(self):
        msg = "read_ports now called num_read_ports for clarity"
        raise PyrtlError(msg)

    def __getitem__(self, addr: WireVectorLike) -> WireVector:
        """Create a read port to read data from the ``MemBlock``.

        :param addr: ``MemBlock`` address to read. A ``WireVector``, or any type that
            can be coerced to ``WireVector`` by :func:`as_wires`.

        :return: A ``WireVector`` containing the data read from the ``MemBlock`` at
                 address ``addr``.
        """
        addr = as_wires(addr, bitwidth=self.addrwidth, truncating=False)
        if len(addr) > self.addrwidth:
            msg = "memory index bitwidth > addrwidth"
            raise PyrtlError(msg)
        return _MemIndexed(mem=self, index=addr)

    def __setitem__(
        self, addr: WireVectorLike, data: MemBlock.EnabledWrite | WireVectorLike
    ):
        """Create a write port to write data to the ``MemBlock``.

        :param addr: ``MemBlock`` address to write. A ``WireVector``, or any type that
            can be coerced to ``WireVector`` by :func:`as_wires`.
        :param data: ``MemBlock`` data to write. An :class:`EnabledWrite`,
            ``WireVector``, or any type that can be coerced to ``WireVector`` by
            :func:`as_wires`.
        """
        if isinstance(data, _MemAssignment):
            self._assignment(addr, data.rhs, is_conditional=data.is_conditional)
        else:
            msg = 'error, assigment to memories should use "<<=" not "=" operator'
            raise PyrtlError(msg)

    def _readaccess(self, addr):
        # FIXME: add conditional read ports
        return self._build_read_port(addr)

    def _build_read_port(self, addr):
        if self.max_read_ports is not None:
            self.num_read_ports += 1
            if self.num_read_ports > self.max_read_ports:
                msg = f"maximum number of read ports ({self.max_read_ports}) exceeded"
                raise PyrtlError(msg)
        data = WireVector(bitwidth=self.bitwidth)
        readport_net = LogicNet(
            op="m", op_param=(self.id, self), args=(addr,), dests=(data,)
        )
        working_block().add_net(readport_net)
        self.readport_nets.append(readport_net)
        return data

    def _assignment(self, item, val, is_conditional):
        from pyrtl.conditional import _build

        # Even though as_wires is already called on item already in the __getitem__
        # method, we need to call it again here because __setitem__ passes the original
        # item to _assignment.
        addr = as_wires(item, bitwidth=self.addrwidth, truncating=False)

        if isinstance(val, MemBlock.EnabledWrite):
            data, enable = val.data, val.enable
        else:
            data, enable = val, Const(1, bitwidth=1)
        data = as_wires(data, bitwidth=self.bitwidth, truncating=False)
        enable = as_wires(enable, bitwidth=1, truncating=False)

        if len(data) != self.bitwidth:
            msg = "error, write data larger than memory bitwidth"
            raise PyrtlError(msg)
        if len(enable) != 1:
            msg = "error, enable signal not exactly 1 bit"
            raise PyrtlError(msg)

        if is_conditional:
            _build(self, (addr, data, enable))
        else:
            self._build(addr, data, enable)

    def _build(self, addr, data, enable):
        """Builds a write port."""
        if self.max_write_ports is not None:
            self.num_write_ports += 1
            if self.num_write_ports > self.max_write_ports:
                msg = f"maximum number of write ports ({self.max_write_ports}) exceeded"
                raise PyrtlError(msg)
        writeport_net = LogicNet(
            op="@", op_param=(self.id, self), args=(addr, data, enable), dests=()
        )
        working_block().add_net(writeport_net)
        self.writeport_nets.append(writeport_net)

    def _make_copy(self, block=None):
        block = working_block(block)
        return MemBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            name=self.name,
            max_read_ports=self.max_read_ports,
            max_write_ports=self.max_write_ports,
            asynchronous=self.asynchronous,
            block=block,
        )


class RomBlock(MemBlock):
    """PyRTL Read Only Memory (ROM).

    ``RomBlocks`` are PyRTL's read only memory block. They support the same read
    interface as :class:`MemBlock`, but they cannot be written to (i.e. there are no
    write ports). The ROM's contents are specified when the ROM is constructed.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example that creates and reads a 4-element ROM::

        >>> rom = pyrtl.RomBlock(bitwidth=3, addrwidth=2, romdata=[4, 5, 6, 7])
        >>> read_addr = pyrtl.Register(name="read_addr", bitwidth=2)
        >>> read_addr.next <<= read_addr + 1
        >>> data = pyrtl.Output(name="data")
        >>> data <<= rom[read_addr]

        >>> sim = pyrtl.Simulation()
        >>> sim.step_multiple(nsteps=6)
        >>> sim.tracer.trace["read_addr"]
        [0, 1, 2, 3, 0, 1]
        >>> sim.tracer.trace["data"]
        [4, 5, 6, 7, 4, 5]
    """

    def __init__(
        self,
        bitwidth: int,
        addrwidth: int,
        romdata,
        name: str = "",
        max_read_ports: int = 2,
        build_new_roms: bool = False,
        asynchronous: bool = False,
        pad_with_zeros: bool = False,
        block: Block = None,
    ):
        """Create a PyRTL Read Only Memory.

        :param bitwidth: The bitwidth of each element in the ROM.
        :param addrwidth: The number of bits used to address an element in the ROM. The
            ROM can store ``2 ** addrwidth`` elements.
        :param romdata: Specifies the data stored in the ROM. This can either be a
            function or an array (iterable) that maps from address to data.
        :param name: The identifier for the memory.
        :param max_read_ports: Limits the number of read ports each block can create;
            passing ``None`` indicates there is no limit.
        :param build_new_roms: Indicates whether :meth:`RomBlock.__getitem__` should
            create copies of the ``RomBlock`` to avoid exceeding ``max_read_ports``.
        :param asynchronous: If ``False``, ensure that all ``RomBlock`` inputs are
            registers, inputs, or constants. See :ref:`asynchronous_memories`.
        :param pad_with_zeros: If ``True``, fill any missing ``romdata`` with zeros so
            all accesses to the ROM are well defined. Otherwise, :class:`Simulation`
            will raise an exception when accessing unintialized data. If you are
            generating Verilog, you will need to specify a value for every address (in
            which case setting this to ``True`` will help), however for testing and
            simulation it useful to know if you are accessing an unspecified value
            (which is why it is ``False`` by default).
        :param block: The block to add to, defaults to the :ref:`working_block`.
        """

        super().__init__(
            bitwidth=bitwidth,
            addrwidth=addrwidth,
            name=name,
            max_read_ports=max_read_ports,
            max_write_ports=0,
            asynchronous=asynchronous,
            block=block,
        )
        self.data = romdata
        self.build_new_roms = build_new_roms
        self.current_copy = self
        self.pad_with_zeros = pad_with_zeros

    def __getitem__(self, addr: WireVector) -> WireVector:
        """Create a read port to read data from the ``RomBlock``.

        If ``build_new_roms`` was specified, create a new copy of the ``RomBlock`` if
        the number of read ports exceeds ``max_read_ports``.

        :param addr: ``MemBlock`` address to read.

        :raises PyrtlError: If ``addr`` is an ``int``. ``RomBlocks`` hold constant data,
            so they don't need to be read when the read address is statically known.
            Create a :class:`Const` with the data at the read address instead.

        :return: A ``WireVector`` containing the data read from the ``RomBlock`` at
                 address ``addr``.
        """
        if isinstance(addr, numbers.Number):
            msg = (
                "There is no point in indexing into a RomBlock with an int. Instead, "
                "get the value from the source data for this Rom"
            )
            raise PyrtlError(msg)
            # If you really know what you are doing, use a Const WireVector instead.
        return super().__getitem__(addr)

    def __setitem__(self, item, assignment):
        msg = "no writing to a read-only memory"
        raise PyrtlError(msg)

    def _get_read_data(self, address: int):
        """_get_read_data is called by the simulator to fetch RomBlock data.

        :param address: address is a dynamic run-time value (an integer), *not* a
            WireVector.

        """
        try:
            if address < 0 or address > 2**self.addrwidth - 1:
                raise PyrtlError("Invalid address, " + str(address) + " specified")
        except TypeError as exc:
            msg = f"Address: {address} with invalid type specified"
            raise PyrtlError(msg) from exc
        if isinstance(self.data, types.FunctionType):
            try:
                value = self.data(address)
            except Exception as exc:
                msg = "Invalid data function for RomBlock"
                raise PyrtlError(msg) from exc
        else:
            try:
                value = self.data[address]
            except KeyError as exc:
                if self.pad_with_zeros:
                    value = 0
                else:
                    msg = (
                        f"RomBlock key {address} is invalid, consider using "
                        "pad_with_zeros=True for defaults"
                    )
                    raise PyrtlError(msg) from exc
            except IndexError as exc:
                if self.pad_with_zeros:
                    value = 0
                else:
                    msg = (
                        f"RomBlock index {address} is invalid, consider using "
                        "pad_with_zeros=True for defaults"
                    )
                    raise PyrtlError(msg) from exc
            except Exception as exc:
                msg = "invalid type for RomBlock data object"
                raise PyrtlError(msg) from exc

        try:
            value = infer_val_and_bitwidth(value, bitwidth=self.bitwidth).value
        except TypeError as exc:
            msg = f"Value: {value} from rom {self} has an invalid type"
            raise PyrtlError(msg) from exc
        return value

    def _build_read_port(self, addr):
        if self.build_new_roms and (
            self.current_copy.num_read_ports >= self.current_copy.max_read_ports
        ):
            self.current_copy = self._make_copy()
        return super(RomBlock, self.current_copy)._build_read_port(addr)

    def _make_copy(
        self,
        block=None,
    ):
        block = working_block(block)
        return RomBlock(
            bitwidth=self.bitwidth,
            addrwidth=self.addrwidth,
            romdata=self.data,
            name=self.name,
            max_read_ports=self.max_read_ports,
            asynchronous=self.asynchronous,
            pad_with_zeros=self.pad_with_zeros,
            block=block,
        )



"""
Basic integer multiplication is defined in PyRTL's core library, see:

- :meth:`.WireVector.__mul__` for unsigned integer multiplication.

- :func:`.signed_mult` for signed integer multiplication.

The functions below provide more complex alternatives.
"""

import math
from typing import Callable

import pyrtl
from pyrtl.rtllib import adders


def simple_mult(
    A: pyrtl.WireVector, B: pyrtl.WireVector, start: pyrtl.WireVector
) -> tuple[pyrtl.Register, pyrtl.WireVector]:
    """Builds a slow, small multiplier using the simple shift-and-add algorithm.

    Requires very small area (it uses only a single adder), but has long delay (worst
    case is ``len(A)`` cycles).

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> start = pyrtl.Input(name="start", bitwidth=1)

        >>> output, done = pyrtl.rtllib.multipliers.simple_mult(a, b, start=start)
        >>> output.name = "output"
        >>> done.name = "done"

        >>> sim = pyrtl.Simulation()
        >>> sim.step({"a": 2, "b": 3, "start": True})
        >>> while not sim.inspect("done"):
        ...     sim.step({"a": 0, "b": 0, "start": False})
        >>> sim.inspect("output")
        6

    :param A: Input wire for the multiplication.
    :param B: Input wire for the multiplication.
    :param start: A one-bit input that indicates when the inputs are ready.

    :return: A :class:`.Register` containing the product, and a 1-bit ``done`` signal.
    """
    triv_result = _trivial_mult(A, B)
    if triv_result is not None:
        return triv_result, pyrtl.Const(1, 1)

    alen = len(A)
    blen = len(B)
    areg = pyrtl.Register(alen)
    breg = pyrtl.Register(blen + alen)
    accum = pyrtl.Register(blen + alen)
    done = pyrtl.WireVector(bitwidth=1)

    # During multiplication, shift a right every cycle, b left every cycle
    with pyrtl.conditional_assignment:
        with start:  # initialization
            areg.next |= A
            breg.next |= B
            accum.next |= 0
        with areg != 0:  # don't run when there's no work to do
            areg.next |= areg[1:]  # right shift
            breg.next |= pyrtl.concat(breg, pyrtl.Const(0, 1))  # left shift
            a_0_val = areg[0].sign_extended(len(accum))

            # adds to accum only when LSB of areg is 1
            accum.next |= accum + (a_0_val & breg)
        with pyrtl.otherwise:
            done |= True

    return accum, done


def _trivial_mult(A, B):
    """
    Turns a multiplication into an And gate if one of the wires is a bitwidth of 1.

    :param A:
    :param B:

    :return:
    """
    if len(B) == 1:
        A, B = B, A  # so that we can reuse the code below :)
    if len(A) == 1:
        a_vals = A.sign_extended(len(B))

        # keep the wirevector len consistent
        return pyrtl.concat_list([a_vals & B, pyrtl.Const(0)])
    return None


def complex_mult(
    A: pyrtl.WireVector, B: pyrtl.WireVector, shifts: int, start: pyrtl.WireVector
) -> tuple[pyrtl.Register, pyrtl.WireVector]:
    """Generate shift-and-add multiplier that can shift and add multiple bits per clock
    cycle. Uses substantially more space than :func:`simple_mult` but is much faster.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> start = pyrtl.Input(name="start", bitwidth=1)

        >>> output, done = pyrtl.rtllib.multipliers.complex_mult(
        ...     a, b, shifts=2, start=start)
        >>> output.name = "output"
        >>> done.name = "done"

        >>> sim = pyrtl.Simulation()
        >>> sim.step({"a": 2, "b": 3, "start": True})
        >>> while not sim.inspect("done"):
        ...     sim.step({"a": 0, "b": 0, "start": False})
        >>> sim.inspect("output")
        6

    :param A: Input wire for the multiplication.
    :param B: Input wire for the multiplication.
    :param shifts: Number of spaces :class:`.Register` is to be shifted per clock cycle.
        Cannot be greater than the length of ``A`` or ``B``.
    :param start: One-bit start signal.

    :return: :class:`.Register` containing the product, and a 1-bit ``done`` signal.
    """

    alen = len(A)
    blen = len(B)
    areg = pyrtl.Register(alen)
    breg = pyrtl.Register(alen + blen)
    accum = pyrtl.Register(alen + blen)
    done = pyrtl.WireVector(bitwidth=1)
    if (shifts > alen) or (shifts > blen):
        msg = (
            "shift is larger than one or both of the parameters A or B, please choose "
            "smaller shift"
        )
        raise pyrtl.PyrtlError(msg)

    # During multiplication, shift a right every cycle 'shift' times, shift b left every
    # cycle 'shift' times
    with pyrtl.conditional_assignment:
        with start:  # initialization
            areg.next |= A
            breg.next |= B
            accum.next |= 0

        with areg != 0:  # don't run when there's no work to do
            # "Multiply" shifted breg by LSB of areg by cond. adding
            areg.next |= pyrtl.shift_right_logical(areg, shifts)
            breg.next |= pyrtl.shift_left_logical(breg, shifts)
            accum.next |= accum + _one_cycle_mult(areg, breg, shifts)

        with pyrtl.otherwise:
            done |= True

    return accum, done


def _one_cycle_mult(areg, breg, rem_bits, sum_sf=0, curr_bit=0):
    """Returns a WireVector sum of ``rem_bits`` multiplies (in one clock cycle)

    ..note::

        this method requires a lot of area because of the indexing in the else statement
    """
    if rem_bits == 0:
        return sum_sf
    a_curr_val = areg[curr_bit].sign_extended(len(breg))
    if curr_bit == 0:  # if no shift
        return _one_cycle_mult(
            areg,
            breg,
            rem_bits - 1,  # areg, breg, rem_bits
            sum_sf + (a_curr_val & breg),  # sum_sf
            curr_bit + 1,
        )  # curr_bit
    return _one_cycle_mult(
        areg,
        breg,
        rem_bits - 1,  # areg, breg, rem_bits
        sum_sf + (a_curr_val & pyrtl.concat(breg, pyrtl.Const(0, curr_bit))),  # sum_sf
        curr_bit + 1,  # curr_bit
    )


def tree_multiplier(
    A: pyrtl.WireVector,
    B: pyrtl.WireVector,
    reducer: Callable = adders.wallace_reducer,
    adder_func: Callable = adders.kogge_stone,
) -> pyrtl.WireVector:
    """Build an fast unclocked multiplier using a Wallace or Dada Tree.

    Delay is `O(log(N))`, while area is `O(N^2)`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.multipliers.tree_multiplier(a, b)

        >>> sim = pyrtl.Simulation()
        >>> sim.step({"a": 2, "b": 3})
        >>> sim.inspect("output")
        6

    :param A: Input wire for the multiplication.
    :param B: Input wire for the multiplication.
    :param reducer: Reducing the tree with a :func:`~.adders.wallace_reducer` or a
        :func:`~.adders.dada_reducer` determines whether the ``tree_multiplier`` is a
        Wallace tree multiplier or a Dada tree multiplier.
    :param adder_func: An adder function that will be used to do the last addition.

    :return: The multiplied result.
    """
    # The two tree multipliers basically works by splitting the multiplication into a
    # series of many additions, and it works by applying 'reductions'.
    triv_res = _trivial_mult(A, B)
    if triv_res is not None:
        return triv_res

    bits_length = len(A) + len(B)

    # create a list of lists, with slots for all the weights (bit-positions)
    bits = [[] for weight in range(bits_length)]

    # AND every bit of A with every bit of B (N^2 results) and store by "weight"
    # (bit-position)
    for i, a in enumerate(A):
        for j, b in enumerate(B):
            bits[i + j].append(a & b)

    return reducer(bits, bits_length, adder_func)


def signed_tree_multiplier(
    A, B, reducer=adders.wallace_reducer, adder_func=adders.kogge_stone
):
    """Same as :func:`tree_multiplier`, but uses two's-complement signed integers.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.multipliers.signed_tree_multiplier(a, b)

        >>> sim = pyrtl.Simulation()
        >>> sim.step({"a": -2, "b": 3})
        >>> pyrtl.val_to_signed_integer(sim.inspect("output"), bitwidth=output.bitwidth)
        -6
    """
    if len(A) == 1 or len(B) == 1:
        msg = "sign bit required, one or both wires too small"
        raise pyrtl.PyrtlError(msg)

    aneg, bneg = A[-1], B[-1]
    a = _twos_comp_conditional(A, aneg)
    b = _twos_comp_conditional(B, bneg)

    res = tree_multiplier(
        a[:-1], b[:-1], reducer=reducer, adder_func=adder_func
    ).zero_extended(len(A) + len(B))
    return _twos_comp_conditional(res, aneg ^ bneg)


def _twos_comp_conditional(
    orig_wire: pyrtl.WireVector, sign_bit: pyrtl.WireVector
) -> pyrtl.WireVector:
    """Returns two's complement of ``orig_wire`` if ``sign_bit`` == 1"""
    return pyrtl.select(sign_bit, (~orig_wire + 1).truncate(len(orig_wire)), orig_wire)


def fused_multiply_adder(
    mult_A: pyrtl.WireVector,
    mult_B: pyrtl.WireVector,
    add: pyrtl.WireVector,
    signed: bool = False,
    reducer: Callable = adders.wallace_reducer,
    adder_func: Callable = adders.kogge_stone,
) -> pyrtl.WireVector:
    """Generate efficient hardware for ``mult_A * mult_B + add``.

    Multiplies two :class:`WireVectors<.WireVector>` together and adds a third
    :class:`.WireVector` to the multiplication result, all in one step. By combining
    these operations, rather than doing them separately, one reduces both the area and
    the timing delay of the circuit.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> c = pyrtl.Input(name="c", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.multipliers.fused_multiply_adder(a, b, c)

        >>> sim = pyrtl.Simulation()
        >>> sim.step({"a": 2, "b": 3, "c": 4})
        >>> pyrtl.val_to_signed_integer(sim.inspect("output"), bitwidth=output.bitwidth)
        10
        >>> 2 * 3 + 4
        10

    :param mult_A: Input wire for the multiplication.
    :param mult_B: Input wire for the multiplication.
    :param add: Input wire for the addition.
    :param signed: Currently not supported (will be added in the future) The default
        will likely be changed to ``True``, so if you want the smallest set of wires in
        the future, specify this as ``False``.
    :param reducer: (advanced) The tree reducer to use. See
        :func:`~.adders.dada_reducer` and :func:`~.adders.wallace_reducer`.
    :param adder_func: (advanced) The adder to use to add the two results at the end.

    :return: The result :class:`.WireVector`.
    """

    # TODO: Specify the length of the result wirevector

    return generalized_fma(((mult_A, mult_B),), (add,), signed, reducer, adder_func)


def generalized_fma(
    mult_pairs: list[tuple[pyrtl.WireVector, pyrtl.WireVector]],
    add_wires: list[pyrtl.WireVector],
    signed: bool = False,  # noqa: ARG001
    reducer: Callable = adders.wallace_reducer,
    adder_func: Callable = adders.kogge_stone,
) -> pyrtl.WireVector:
    """Generated an optimized fused multiply adder.

    A generalized FMA unit that multiplies each pair of numbers in ``mult_pairs``, then
    adds up the resulting products and all the values of the ``add_wires``. This is
    faster than multiplying and adding separately because you avoid unnecessary adder
    structures for intermediate representations.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> mult_pairs = [(pyrtl.Const(2), pyrtl.Const(3)),
        ...               (pyrtl.Const(4), pyrtl.Const(5))]
        >>> add_wires = [pyrtl.Const(6), pyrtl.Const(7)]
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtllib.multipliers.generalized_fma(mult_pairs, add_wires)

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> sim.inspect("output")
        39
        >>> 2 * 3 + 4 * 5 + 6 + 7
        39

    :param mult_pairs: Either ``None`` (if there are no pairs to multiply) or a list of
        pairs of wires to multiply together: ``[(mult1_1, mult1_2), ...]``
    :param add_wires: Either ``None`` (if there are no individual items to add other
        than the ``mult_pairs`` products), or a list of wires to add on top of the
        result of the pair multiplication.
    :param signed: Currently not supported (will be added in the future) The default
        will likely be changed to ``True``, so if you want the smallest set of wires in
        the future, specify this as ``False``.
    :param reducer: (advanced) The tree reducer to use. See
        :func:`~.adders.dada_reducer` and :func:`~.adders.wallace_reducer`.
    :param adder_func: (advanced) The adder to use to add the two results at the end.

    :return: The result :class:`.WireVector`.
    """
    # first need to figure out the max length
    if mult_pairs:  # Need to deal with the case when it is empty
        mult_max = max(len(m[0]) + len(m[1]) - 1 for m in mult_pairs)
    else:
        mult_max = 0

    if add_wires:
        add_max = max(len(x) for x in add_wires)
    else:
        add_max = 0

    longest_wire_len = max(add_max, mult_max)
    bits = [[] for i in range(longest_wire_len)]

    for mult_a, mult_b in mult_pairs:
        for i, a in enumerate(mult_a):
            for j, b in enumerate(mult_b):
                bits[i + j].append(a & b)

    for wire in add_wires:
        for bit_loc, bit in enumerate(wire):
            bits[bit_loc].append(bit)

    result_bitwidth = longest_wire_len + math.ceil(
        math.log2(len(add_wires) + len(mult_pairs))
    )
    return reducer(bits, result_bitwidth, adder_func)



"""Some useful hardware generators (e.g. muxes, signed multipliers, etc.)"""

from __future__ import annotations

import itertools

from pyrtl.conditional import otherwise
from pyrtl.core import Block, LogicNet, working_block
from pyrtl.pyrtlexceptions import PyrtlError, PyrtlInternalError
from pyrtl.rtllib import barrel, muxes
from pyrtl.wire import Const, WireVector, WireVectorLike, WrappedWireVector


def mux(
    index: WireVectorLike, *mux_ins: WireVectorLike, default: WireVectorLike = None
) -> WireVector:
    """Multiplexer returning a wire from ``mux_ins`` according to ``index``.

    ``index`` ``0`` corresponds to the first ``mux_in`` argument.

    .. note::

        If ``index`` is a 1-bit predicate (something that is ``True`` or ``False``
        rather than an integer), it is clearer to use :func:`select`, whose argument
        order is consistent with the ternary operators in C-style languages
        (``condition ? truecase : falsecase``).

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example of multiplexing between four values::

        >>> index = pyrtl.Input(name="index", bitwidth=2)
        >>> selected = pyrtl.WireVector(name="selected")

        >>> selected <<= pyrtl.mux(index, 4, 5, 6, 7)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"index": 0})
        >>> sim.inspect("selected")
        4

        >>> sim.step(provided_inputs={"index": 3})
        >>> sim.inspect("selected")
        7

    .. doctest only::

        >>> pyrtl.reset_working_block()

    Example with ``default``::

        >>> index = pyrtl.Input(name="index", bitwidth=2)
        >>> selected = pyrtl.WireVector(name="selected")
        >>> selected <<= pyrtl.mux(index, 4, 5, default=9)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"index": 1})
        >>> sim.inspect("selected")
        5

        >>> sim.step(provided_inputs={"index": 2})
        >>> sim.inspect("selected")
        9

    :param index: Multiplexer's selection input. Can be a :class:`WireVector`, or any
        type that can be coerced to :class:`WireVector` by :func:`as_wires`.
    :param mux_ins: :class:`WireVector` arguments to select from. Can be a
        :class:`WireVector`, or any type that can be coerced to :class:`WireVector` by
        :func:`as_wires`.
    :param default: If you are selecting between fewer items than ``index`` can address,
        ``default`` will be used for all remaining items. For example, if you have a
        3-bit index but are selecting between 6 ``mux_ins``, you need to specify a value
        for those other 2 possible values of ``index`` (``0b110`` and ``0b111``).

    :raises PyrtlError: If there are not enough ``mux_ins`` to select from. If
        ``default=None``, the number of ``mux_ins`` must be exactly ``2 **
        index.bitwidth``.

    :return: :class:`WireVector` with :attr:`~WireVector.bitwidth` matching the length
             of the longest input (not including ``index``).
    """
    # find the diff between the addressable range and number of inputs given
    index = as_wires(index)
    short_by = 2 ** len(index) - len(mux_ins)
    if short_by > 0 and default is not None:  # extend the list to appropriate size
        mux_ins = list(mux_ins)
        mux_ins.extend([default] * short_by)

    if 2 ** len(index) != len(mux_ins):
        msg = (
            f"Mux select line is {len(index)} bits, but selecting from {len(mux_ins)} "
            "inputs."
        )
        raise PyrtlError(msg)

    if len(index) == 1:
        return select(index, falsecase=mux_ins[0], truecase=mux_ins[1])
    half = len(mux_ins) // 2
    return select(
        index[-1],
        falsecase=mux(index[0:-1], *mux_ins[:half]),
        truecase=mux(index[0:-1], *mux_ins[half:]),
    )


def select(
    sel: WireVectorLike, truecase: WireVectorLike, falsecase: WireVectorLike
) -> WireVector:
    """Multiplexer returning ``truecase`` when ``sel == 1``, otherwise ``falsecase``.

    ``select`` is equivalent to :func:`mux` with a 1-bit ``index``, except that
    ``select``'s ``truecase`` is its first argument, rather than its second.
    ``select``'s argument order is consistent with the ternary operator in C-style
    languages, which improves readability, so prefer ``select`` over :func:`mux` when
    selecting between exactly two options.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example that computes ``min(a, 5)``::

        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> min = pyrtl.WireVector(name="min")

        >>> min <<= pyrtl.select(a < 5, 5, a)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"a": 4})
        >>> sim.inspect("min")
        5

        >>> sim.step(provided_inputs={"a": 6})
        >>> sim.inspect("min")
        6

    :param sel: Multiplexer's selection input. Can be a :class:`WireVector`, or any type
        that can be coerced to :class:`WireVector` by :func:`as_wires`.
    :param truecase: The WireVector selected if ``sel == 1``. Can be a
        :class:`WireVector`, or any type that can be coerced to :class:`WireVector` by
        :func:`as_wires`.
    :param falsecase: The WireVector selected if ``sel == 0``. Can be a
        :class:`WireVector`, or any type that can be coerced to :class:`WireVector` by
        :func:`as_wires`.

    :return: :class:`WireVector` with :attr:`~WireVector.bitwidth` matching the longer
             of ``truecase`` and ``falsecase``.
    """
    sel, f, t = (as_wires(w) for w in (sel, falsecase, truecase))
    f, t = match_bitwidth(f, t)
    outwire = WireVector(bitwidth=len(f))

    net = LogicNet(op="x", op_param=None, args=(sel, f, t), dests=(outwire,))
    working_block().add_net(net)  # this includes sanity check on the mux
    return outwire


def concat(*args: WireVectorLike) -> WireVector:
    """Concatenates multiple :class:`WireVectors<WireVector>` into a single
    :class:`WireVector`.

    Concatenates any number of arguments. The right-most argument is the least
    significant bits of the result.

    .. note::

        If you have a :class:`list` of arguments to ``concat`` together, you probably
        want index 0 to be the least significant bit. If so, and you unpack the
        :class:`list` into ``concat``'s ``args``, the result will be backwards. The
        function :func:`concat_list` is provided specifically for that case.

    .. note::

        Consider using :func:`wire_struct` or :func:`wire_matrix` instead, which helps
        with consistently disassembling, naming, and reassembling fields.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example that concatenates two bytes into a 16-bit ``output``::

        >>> msb = pyrtl.Input(name="msb", bitwidth=8)
        >>> lsb = pyrtl.Input(name="lsb", bitwidth=8)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.concat(msb, lsb)
        >>> output.bitwidth
        16

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"msb": 0xab, "lsb": 0xcd})
        >>> hex(sim.inspect("output"))
        '0xabcd'

    :param args: Inputs to concatenate, with the most significant bits first. Each input
        can be a :class:`WireVector`, or any type that can be coerced to
        :class:`WireVector` by :func:`as_wires`.

    :return: :class:`WireVector` with :attr:`~WireVector.bitwidth` equal to the sum of
             all ``args``' :attr:`bitwidths<~WireVector.bitwidth>`.
    """
    if len(args) <= 0:
        msg = "error, concat requires at least 1 argument"
        raise PyrtlError(msg)
    if len(args) == 1:
        return as_wires(args[0])

    arg_wirevectors = tuple(as_wires(arg) for arg in args)
    final_width = sum(len(arg) for arg in arg_wirevectors)
    outwire = WireVector(bitwidth=final_width)
    net = LogicNet(op="c", op_param=None, args=arg_wirevectors, dests=(outwire,))
    working_block().add_net(net)
    return outwire


def concat_list(wire_list: list[WireVectorLike]) -> WireVector:
    """Concatenates a list of :class:`WireVectors<WireVector>` into a single
    :class:`WireVector`.

    This take a :class:`list` of :class:`WireVectors<WireVector>` and concats them all
    into a single :class:`WireVector`, with the element at index 0 serving as the least
    significant bits. This is useful when you have a variable number of
    :class:`WireVectors<WireVector>` to concatenate, otherwise :func:`concat` is
    prefered.

    .. note::

        Consider using :func:`wire_struct` or :func:`wire_matrix` instead, which helps
        with consistently disassembling, naming, and reassembling fields.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example that concatenates two bytes into a 16-bit ``output``::

        >>> msb = pyrtl.Input(name="msb", bitwidth=8)
        >>> lsb = pyrtl.Input(name="lsb", bitwidth=8)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.concat_list([lsb, msb])
        >>> output.bitwidth
        16

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"msb": 0xab, "lsb": 0xcd})
        >>> hex(sim.inspect("output"))
        '0xabcd'

    :param wire_list: List of inputs to concatenate. Each input can be a
        :class:`WireVector`, or any type that can be coerced to :class:`WireVector` by
        :func:`as_wires`.

    :return: :class:`WireVector` with :attr:`~WireVector.bitwidth` equal to the sum of
             all ``wire_list`` :attr:`bitwidths<~WireVector.bitwidth>`.
    """
    return concat(*reversed(wire_list))


def _signed_input_to_wirevector(x):
    """Convert int input to a signed Const, otherwise call `as_wires`."""
    if isinstance(x, int):
        return Const(x, signed=True)
    return as_wires(x)


def signed_add(a: WireVectorLike, b: WireVectorLike) -> WireVector:
    """Return a :class:`WireVector` for the result of signed addition.

    The inputs are :meth:`~WireVector.sign_extended` to the result's bitwidth before
    adding.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> neg_three = pyrtl.Const(val=-3, signed=True, bitwidth=3)
        >>> neg_five = pyrtl.Const(val=-5, signed=True, bitwidth=5)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.signed_add(neg_three, neg_five)
        >>> output.bitwidth
        6

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> pyrtl.val_to_signed_integer(sim.inspect("output"), bitwidth=output.bitwidth)
        -8

    :param a: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.
    :param b: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.

    :return: A :class:`WireVector` representing the sum of ``a`` and ``b``, with
             :attr:`~WireVector.bitwidth` ``max(a.bitwidth, b.bitwidth) + 1``.
    """
    a = _signed_input_to_wirevector(a)
    b = _signed_input_to_wirevector(b)
    result_bitwidth = max(a.bitwidth, b.bitwidth) + 1
    a = a.sign_extended(result_bitwidth)
    b = b.sign_extended(result_bitwidth)
    return (a + b).truncate(result_bitwidth)


def signed_sub(a: WireVectorLike, b: WireVectorLike) -> WireVector:
    """Return a :class:`WireVector` for the result of signed subtraction.

    The inputs are :meth:`~WireVector.sign_extended` to the result's bitwidth before
    subtracting.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> neg_three = pyrtl.Const(val=-3, signed=True, bitwidth=3)
        >>> neg_five = pyrtl.Const(val=-5, signed=True, bitwidth=5)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.signed_sub(neg_three, neg_five)
        >>> output.bitwidth
        6

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> pyrtl.val_to_signed_integer(sim.inspect("output"), bitwidth=output.bitwidth)
        2

    :param a: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.
    :param b: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.

    :return: A :class:`WireVector` representing the difference between ``a`` and ``b``,
             with :attr:`~WireVector.bitwidth` ``max(a.bitwidth, b.bitwidth) + 1``.
    """
    a = _signed_input_to_wirevector(a)
    b = _signed_input_to_wirevector(b)
    result_bitwidth = max(a.bitwidth, b.bitwidth) + 1
    a = a.sign_extended(result_bitwidth)
    b = b.sign_extended(result_bitwidth)
    return (a - b).truncate(result_bitwidth)


def mult_signed(a, b):
    """mult_signed is now deprecated, use ``signed_mult`` instead"""
    return signed_mult(a, b)


def signed_mult(a: WireVectorLike, b: WireVectorLike) -> WireVector:
    """Return a :class:`WireVector` for the result of signed multiplication.

    The inputs are :meth:`~WireVector.sign_extended` to the result's bitwidth before
    multiplying.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> neg_three = pyrtl.Const(val=-3, signed=True, bitwidth=3)
        >>> neg_five = pyrtl.Const(val=-5, signed=True, bitwidth=5)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.signed_mult(neg_three, neg_five)
        >>> output.bitwidth
        8

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> pyrtl.val_to_signed_integer(sim.inspect("output"), bitwidth=output.bitwidth)
        15

    :param a: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.
    :param b: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.

    :return: A :class:`WireVector` representing the product of ``a`` and ``b``, with
             :attr:`~WireVector.bitwidth` ``a.bitwidth + b.bitwidth``.
    """
    a = _signed_input_to_wirevector(a)
    b = _signed_input_to_wirevector(b)
    result_bitwidth = a.bitwidth + b.bitwidth
    a = a.sign_extended(result_bitwidth)
    b = b.sign_extended(result_bitwidth)
    return (a * b).truncate(result_bitwidth)


def signed_lt(a: WireVectorLike, b: WireVectorLike) -> WireVector:
    """Return a 1-bit :class:`WireVector` for the result of a signed ``<`` comparison.

    The inputs are :meth:`~WireVector.sign_extended` to matching bitwidths before
    comparing.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> neg_three = pyrtl.Const(val=-3, signed=True, bitwidth=3)
        >>> neg_five = pyrtl.Const(val=-5, signed=True, bitwidth=5)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.signed_lt(neg_three, neg_five)
        >>> output.bitwidth
        1

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> sim.inspect("output")
        0

    :param a: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.
    :param b: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.

    :return: A 1-bit :class:`WireVector` indicating if ``a`` is less than ``b``.
    """
    a, b = match_bitwidth(as_wires(a), as_wires(b), signed=True)
    r = a - b
    return r[-1] ^ (~a[-1]) ^ (~b[-1])


def signed_le(a: WireVectorLike, b: WireVectorLike) -> WireVector:
    """Return a 1-bit :class:`WireVector` for the result of a signed ``<=`` comparison.

    The inputs are :meth:`~WireVector.sign_extended` to matching bitwidths before
    comparing.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> neg_three = pyrtl.Const(val=-3, signed=True, bitwidth=3)
        >>> neg_five = pyrtl.Const(val=-5, signed=True, bitwidth=5)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.signed_le(neg_three, neg_five)
        >>> output.bitwidth
        1

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> sim.inspect("output")
        0

    :param a: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.
    :param b: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.

    :return: A 1-bit :class:`WireVector` indicating if ``a`` is less than or equal to
             ``b``.
    """
    a, b = match_bitwidth(as_wires(a), as_wires(b), signed=True)
    r = a - b
    return (r[-1] ^ (~a[-1]) ^ (~b[-1])) | (a == b)


def signed_gt(a: WireVectorLike, b: WireVectorLike) -> WireVector:
    """Return a 1-bit :class:`WireVector` for the result of a signed ``>`` comparison.

    The inputs are :meth:`~WireVector.sign_extended` to matching bitwidths before
    comparing.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> neg_three = pyrtl.Const(val=-3, signed=True, bitwidth=3)
        >>> neg_five = pyrtl.Const(val=-5, signed=True, bitwidth=5)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.signed_gt(neg_three, neg_five)
        >>> output.bitwidth
        1

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> sim.inspect("output")
        1

    :param a: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.
    :param b: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.

    :return: A 1-bit :class:`WireVector` indicating if ``a`` is greater than ``b``.
    """
    a, b = match_bitwidth(as_wires(a), as_wires(b), signed=True)
    r = b - a
    return r[-1] ^ (~a[-1]) ^ (~b[-1])


def signed_ge(a: WireVectorLike, b: WireVectorLike) -> WireVector:
    """Return a 1-bit :class:`WireVector` for the result of a signed ``>=`` comparison.

    The inputs are :meth:`~WireVector.sign_extended` to matching bitwidths before
    comparing.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> neg_three = pyrtl.Const(val=-3, signed=True, bitwidth=3)
        >>> neg_five = pyrtl.Const(val=-5, signed=True, bitwidth=5)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.signed_ge(neg_three, neg_five)
        >>> output.bitwidth
        1

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> sim.inspect("output")
        1

    :param a: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.
    :param b: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`.

    :return: A 1-bit :class:`WireVector` indicating if ``a`` is greater than or equal to
             ``b``.
    """
    a, b = match_bitwidth(as_wires(a), as_wires(b), signed=True)
    r = b - a
    return (r[-1] ^ (~a[-1]) ^ (~b[-1])) | (a == b)


def shift_right_arithmetic(
    bits_to_shift: WireVector, shift_amount: WireVector | int
) -> WireVector:
    """Arithmetic right shift operation.

    Arithemetic shifting treats the ``bits_to_shift`` as a signed number, so copies of
    ``bits_to_shift``'s sign bit will be added on the left.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> neg_forty = pyrtl.Const(val=-40, signed=True, bitwidth=7)
        >>> shift_amount = pyrtl.Input(name="shift_amount", bitwidth=3)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.shift_right_arithmetic(neg_forty, shift_amount)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"shift_amount": 3})
        >>> pyrtl.val_to_signed_integer(sim.inspect("output"), bitwidth=output.bitwidth)
        -5
        >>> int(-40 / 2 ** 3)
        -5

    Right shifting by ``N`` bits is equivalent to dividing by ``2^N``.

    :param bits_to_shift: Value to shift right arithmetically.
    :param shift_amount: Number of bit positions to shift, as an unsigned integer.

    :return: A new :class:`WireVector` with the same bitwidth as ``bits_to_shift``.
    """
    if isinstance(shift_amount, int):
        if shift_amount >= bits_to_shift.bitwidth:
            return bits_to_shift[-1].sign_extended(len(bits_to_shift))

        return bits_to_shift[shift_amount:].sign_extended(len(bits_to_shift))

    bit_in = bits_to_shift[-1]  # shift in sign_bit
    dir = barrel.Direction.RIGHT
    return barrel.barrel_shifter(bits_to_shift, bit_in, dir, shift_amount)


def shift_left_logical(
    bits_to_shift: WireVector, shift_amount: WireVector | int
) -> WireVector:
    """Logical left shift operation.

    Logical shifting treats the ``bits_to_shift`` as an unsigned number. Zeroes will be
    added on the right and the result will be truncated to ``bits_to_shift.bitwidth``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> three = pyrtl.Const(val=3, bitwidth=6)
        >>> shift_amount = pyrtl.Input(name="shift_amount", bitwidth=3)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.shift_left_logical(three, shift_amount)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"shift_amount": 3})
        >>> sim.inspect("output")
        24
        >>> 3 * 2 ** 3
        24

    Left shifting by ``N`` bits is equivalent to multiplying by ``2^N``.

    :param bits_to_shift: Value to shift left logically.
    :param shift_amount: Number of bit positions to shift, as an unsigned integer.

    :return: A new :class:`WireVector` with the same bitwidth as ``bits_to_shift``.
    """
    if isinstance(shift_amount, int):
        if shift_amount >= bits_to_shift.bitwidth:
            return Const(val=0, bitwidth=bits_to_shift.bitwidth)

        return concat(bits_to_shift[:-shift_amount], Const(0, shift_amount))

    bit_in = 0  # shift in a 0
    dir = barrel.Direction.LEFT
    return barrel.barrel_shifter(bits_to_shift, bit_in, dir, shift_amount)


shift_left_arithmetic = shift_left_logical
"""Alias for :func:`shift_left_logical`"""


def shift_right_logical(
    bits_to_shift: WireVector, shift_amount: WireVector | int
) -> WireVector:
    """Logical right shift operation.

    Logical shifting treats the ``bits_to_shift`` as an unsigned number, so zeroes will
    be added on the left.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> forty = pyrtl.Const(val=40, bitwidth=6)
        >>> shift_amount = pyrtl.Input(name="shift_amount", bitwidth=3)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.shift_right_logical(forty, shift_amount)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"shift_amount": 3})
        >>> sim.inspect("output")
        5
        >>> int(40 / 2 ** 3)
        5

    Right shifting by ``N`` bits is equivalent to dividing by ``2^N``.

    :param bits_to_shift: Value to shift right logically.
    :param shift_amount: Number of bit positions to shift, as an unsigned integer.

    :return: A new :class:`WireVector` with the same bitwidth as ``bits_to_shift``.
    """
    if isinstance(shift_amount, int):
        if shift_amount >= bits_to_shift.bitwidth:
            return Const(val=0, bitwidth=bits_to_shift.bitwidth)
        return bits_to_shift[shift_amount:].zero_extended(len(bits_to_shift))

    bit_in = 0  # shift in a 0
    dir = barrel.Direction.RIGHT
    return barrel.barrel_shifter(bits_to_shift, bit_in, dir, shift_amount)


def match_bitwidth(*args: WireVector, signed: bool = False) -> tuple[WireVector]:
    """Matches multiple :class:`WireVector` :attr:`bitwidths<~WireVector.bitwidth>` via
    zero- or sign-extension.

    :class:`WireVectors<WireVector>` with shorter
    :attr:`bitwidths<~WireVector.bitwidth>` will be to match the longest
    :attr:`~WireVector.bitwidth` in ``args``. :class:`WireVectors<WireVector>` will be
    :meth:`~WireVector.sign_extended` or :meth:`~WireVector.zero_extended`, depending on
    ``signed``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example with sign-extension::

        >>> a = pyrtl.Const(-1, name="a_short", signed=True, bitwidth=2)
        >>> b = pyrtl.Const(-3, name="b", signed=True, bitwidth=4)

        >>> a, b = match_bitwidth(a, b, signed=True)
        >>> a.name = "a_long"
        >>> a.bitwidth, b.bitwidth
        (4, 4)

        >>> sim = pyrtl.Simulation()
        >>> sim.step()
        >>> bin(sim.inspect("b"))
        '0b1101'
        >>> bin(sim.inspect("a_short"))
        '0b11'
        >>> bin(sim.inspect("a_long"))
        '0b1111'

    :param args: :class:`WireVectors<WireVector>` of which to match
        :attr:`~WireVector.bitwidth`
    :param signed: If ``True``, extend shorter :class:`WireVectors<WireVector>` with
        :meth:`~WireVector.sign_extended`. Otherwise, extend with
        :meth:`~WireVector.zero_extended`.

    :return: :class:`tuple` of :class:`WireVectors<WireVector>`, in the same order they
             appeared in ``args``, all with :attr:`~WireVector.bitwidth` equal to the
             longest :attr:`~WireVector.bitwidth` in ``args``.
    """
    max_len = max(len(wv) for wv in args)
    if signed:
        return (wv.sign_extended(max_len) for wv in args)
    return (wv.zero_extended(max_len) for wv in args)


def as_wires(
    val: WireVectorLike,
    bitwidth: int | None = None,
    truncating: bool = True,
    block: Block = None,
) -> WireVector:
    """Convert ``val`` to a :class:`WireVector`.

    ``val`` may be a :class:`WireVector`, :class:`int` (including
    :class:`~enum.IntEnum`), :class:`str`, or :class:`bool`.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    ``as_wires`` is mainly used to coerce values into :class:`WireVectors<WireVector>`
    (for example, operations such as ``x + 1`` where ``1`` needs to be converted to a
    :class:`Const` :class:`WireVector`). See :ref:`wirevector_coercion`. An example::

        >>> def make_my_hardware(a, b):
        ...     a = as_wires(a)
        ...     b = as_wires(b)
        ...     return (a + b) & 0xf

        >>> input = pyrtl.Input(name="input", bitwidth=8)
        >>> output = pyrtl.Output(name="output", bitwidth=4)

        >>> output <<= make_my_hardware(input, 7)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 20})
        >>> sim.inspect("output")
        11
        >>> (20 + 7) % 16
        11

    In the example above, ``as_wires`` will convert the ``7`` to ``Const(7)`` but keep
    ``input`` unchanged.

    :param val: A :class:`WireVector`, or a constant value that can be converted into a
        :class:`Const`.
    :param bitwidth: The :attr:`~WireVector.bitwidth` of the resulting
        :class:`WireVector`.
    :param truncating: Determines whether bits will be dropped to achieve the desired
        :attr:`~WireVector.bitwidth` if ``val`` is too long (if ``True``, the
        most-significant bits will be dropped).
    :param block: ``Block`` to use for the returned :class:`WireVector`. Defaults to the
        :ref:`working_block`.
    """
    from pyrtl.memory import _MemIndexed

    block = working_block(block)

    if isinstance(val, (int, str)):
        # note that this case captures bool as well (as bools are instances of ints)
        return Const(val, bitwidth=bitwidth, block=block)
    if isinstance(val, _MemIndexed):
        # convert to a memory read when the value is actually used
        if val.wire is None:
            val.wire = as_wires(
                val.mem._readaccess(val.index), bitwidth, truncating, block
            )
        return val.wire
    if isinstance(val, WrappedWireVector):
        return val.wire
    if not isinstance(val, WireVector):
        msg = (
            "error, expecting a wirevector, int, or Verilog-style const string got "
            f"{val} instead"
        )
        raise PyrtlError(msg)
    if bitwidth == "0":
        msg = "error, bitwidth must be >= 1"
        raise PyrtlError(msg)
    if val.bitwidth is None:
        msg = "error, attempting to use wirevector with no defined bitwidth"
        raise PyrtlError(msg)
    if bitwidth and bitwidth > val.bitwidth:
        return val.zero_extended(bitwidth)
    if bitwidth and truncating and bitwidth < val.bitwidth:
        return val[:bitwidth]  # truncate the upper bits
    return val


def bitfield_update(
    w: WireVectorLike,
    range_start: int,
    range_end: int,
    newvalue: int,
    truncating: bool = False,
) -> WireVector:
    """Update a :class:`WireVector` by replacing some of its bits with ``newvalue``.

    Given a :class:`WireVector` ``w``, this function returns a new :class:`WireVector`
    that is identical to ``w`` except in the range of bits specified by ``[range_start,
    range_end)``. In that range, the value ``newvalue`` is swapped in. For example::

        bitfield_update(w, range_start=20, range_end=23, newvalue=0b111)

    will return a :class:`WireVector` of the same length as ``w``, and with the same
    values as ``w``, but with bits 20, 21, and 22 all set to ``1``.

    Note that ``range_start`` and ``range_end`` will be inputs to a slice and so
    standard Python slicing rules apply (e.g. negative values for end-relative indexing
    and support for ``None``)::

        # Sets bits 20, 21, 22 to 1.
        w = bitfield_update(w, 20, 23, 0b111)

        # Sets bit 20 to 0, bits 21 and 22 to 1.
        w = bitfield_update(w, 20, 23, 0b110)

        # Assuming w is 32 bits, sets bits 31..20 = 0x7.
        w = bitfield_update(w, 20, None, 0x7)

        # Set the MSB (bit) to 1.
        w = bitfield_update(w, -1, None, 0x1)

        # Set the bits before the MSB (bit) to 9.
        w = bitfield_update(w, None, -1, 0x9)

        # Set the LSB (bit) to 1.
        w = bitfield_update(w, None, 1, 0x1)

        # Set the bits after the LSB (bit) to 9.
        w = bitfield_update(w, 1, None, 0x9)

    .. note::

        Consider using :func:`wire_struct` or :func:`wire_matrix` instead, which helps
        with consistently disassembling, naming, and reassembling fields.

    :param w: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`, to use as the starting point for
              the update
    :param range_start: The start of the range of bits to be updated.
    :param range_end: The end of the range of bits to be updated.
    :param newvalue: The value to be written in to the ``range_start:range_end`` range.
    :param truncating: If ``True``, clip ``newvalue`` to the proper bitwidth if
        ``newvalue`` is too large.

    :raise PyrtlError: If ``newvalue`` is too large to fit in the selected range of bits
        and ``truncating`` is ``False``.

    :return: ``w`` with some of the bits overwritten by ``newvalue``.
    """
    w = as_wires(w)
    idxs = list(
        range(len(w))
    )  # we make a list of integers and slice those up to use as indexes
    idxs_middle = idxs[range_start:range_end]
    if len(idxs_middle) == 0:
        msg = "Cannot update bitfield of size 0 (i.e. there are no bits to update)"
        raise PyrtlError(msg)
    idxs_lower = idxs[: idxs_middle[0]]
    idxs_upper = idxs[idxs_middle[-1] + 1 :]

    newvalue = as_wires(newvalue, bitwidth=len(idxs_middle), truncating=truncating)
    if len(idxs_middle) != len(newvalue):
        msg = (
            f"Cannot update bitfield of length {len(idxs_middle)} with value of length "
            f"{len(newvalue)} unless truncating=True is specified"
        )
        raise PyrtlError(msg)

    result_list = []
    if idxs_lower:
        result_list.append(w[idxs_lower[0] : idxs_lower[-1] + 1])
    result_list.append(newvalue)
    if idxs_upper:
        result_list.append(w[idxs_upper[0] : idxs_upper[-1] + 1])
    result = concat_list(result_list)

    if len(result) != len(w):
        msg = f"len(result)={len(result)}, len(original)={len(w)}"
        raise PyrtlInternalError(msg)
    return result


def bitfield_update_set(
    w: WireVectorLike,
    update_set: dict[tuple[int, int], WireVectorLike],
    truncating: bool = False,
) -> WireVector:
    """Update a :class:`WireVector` by replacing the bits specified in ``update_set``.

    Given a WireVector ``w``, return a new :class:`WireVector` that is identical to `w`
    except in the ranges of bits specified by ``update_set``. When multiple
    non-overlapping fields need to be updated in a single cycle, this provides a clearer
    way to describe that behavior than iterative calls to :func:`bitfield_update`::

        w = bitfield_update_set(w, update_set={
                (20, 23):    0x6,      # sets bit 20 to 0, bits 21 and 22 to 1
                (26, None):  0x7,      # assuming w is 32 bits, sets bits 31..26 to 0x7
                (None, 1):   0x0,      # set the LSB (bit) to 0
            })

    .. note::

        Consider using :func:`wire_struct` or :func:`wire_matrix` instead, which helps
        with consistently disassembling, naming, and reassembling fields.

    :param w: A :class:`WireVector`, or any type that can be coerced to
              :class:`WireVector` by :func:`as_wires`, to use as the starting point for
              the update
    :param update_set: A map from tuples of ``(range_start, range_end)`` integers to a
        new value for the range of bits.
    :param truncating: If ``True``, clip new values to the proper bitwidth if a new
        value is too large.

    :raise PyrtlError: If ``update_set`` contains overlapping fields.

    :return: ``w`` with some of its bits updated.
    """
    w = as_wires(w)
    # keep a list of bits that are updated to find overlaps
    setlist = [False] * len(w)
    # call bitfield for each one
    for (range_start, range_end), new_value in update_set.items():
        # check for overlaps
        setbits = setlist[range_start:range_end]
        if any(setbits):
            msg = "Bitfields for update are overlapping"
            raise PyrtlError(msg)
        setlist[range_start:range_end] = [True] * len(setbits)
        # do the actual update
        w = bitfield_update(w, range_start, range_end, new_value, truncating)
    return w


def enum_mux(
    cntrl: WireVector,
    table: dict[int, WireVector],
    default: WireVector = None,
    strict: bool = True,
) -> WireVector:
    """Build a mux for the control signals specified by an :class:`enum.IntEnum`.

    .. note::

        Consider using :ref:`conditional_assignment` instead.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> from enum import IntEnum
        >>> class Command(IntEnum):
        ...     ADD = 0
        ...     SUB = 1

        >>> command = pyrtl.Input(name="command", bitwidth=1)
        >>> a = pyrtl.Input(name="a", bitwidth=4)
        >>> b = pyrtl.Input(name="b", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.enum_mux(cntrl=command, table={
        ...     Command.ADD: a + b,
        ...     Command.SUB: a - b,
        ... })

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"command": Command.ADD, "a": 1, "b": 2})
        >>> sim.inspect("output")
        3

        >>> sim.step(provided_inputs={"command": Command.SUB, "a": 5, "b": 3})
        >>> sim.inspect("output")
        2

    :param cntrl: Control for the mux.
    :param table: Maps :class:`enum.IntEnum` values to :class:`WireVector`.
    :param default: A :class:`WireVector` to use when the key is not present. In
        addition it is possible to use the key :data:`otherwise` to specify a default
        value, but it is an error if both are supplied.
    :param strict: When ``True``, check that the dictionary has an entry for every
        possible value in the :class:`enum.IntEnum`. Note that if a ``default`` is set,
        then this check is not performed as the ``default`` will provide valid values
        for any underspecified keys.

    :return: Result of the mux.
    """
    # check dictionary keys are of the right type
    keytypeset = {type(x) for x in table if x is not otherwise}
    if len(keytypeset) != 1:
        msg = f"table mixes multiple types {keytypeset} as keys"
        raise PyrtlError(msg)
    keytype = next(iter(keytypeset))
    # check that dictionary is complete for the enum
    try:
        enumkeys = list(keytype.__members__.values())
    except AttributeError as exc:
        msg = f"type {keytype} not an Enum and does not support the same interface"
        raise PyrtlError(msg) from exc
    missingkeys = [e for e in enumkeys if e not in table]

    # check for "otherwise" in table and move it to a default
    if otherwise in table:
        if default is not None:
            msg = 'both "otherwise" and default provided to enum_mux'
            raise PyrtlError(msg)
        default = table[otherwise]

    if strict and default is None and missingkeys:
        msg = f"table provided is incomplete, missing: {missingkeys}"
        raise PyrtlError(msg)

    # generate the actual mux
    vals = {k.value: d for k, d in table.items() if k is not otherwise}
    if default is not None:
        vals["default"] = default
    return muxes.sparse_mux(cntrl, vals)


def and_all_bits(vector: WireVector) -> WireVector:
    """Returns the result of bitwise ANDing all the bits in ``vector``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> input = pyrtl.Input(name="input", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.and_all_bits(input)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0b0101})
        >>> sim.inspect("output")
        0
        >>> sim.step(provided_inputs={"input": 0b1111})
        >>> sim.inspect("output")
        1

    :param vector: Takes a single arbitrary length :class:`WireVector`.

    :return: Returns a 1-bit result, the bitwise ``&`` of all of the bits in ``vector``.
    """
    return tree_reduce(lambda a, b: a & b, vector)


def or_all_bits(vector: WireVector) -> WireVector:
    """Returns the result of bitwise ORing all the bits in ``vector``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> input = pyrtl.Input(name="input", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.or_all_bits(input)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0b0000})
        >>> sim.inspect("output")
        0
        >>> sim.step(provided_inputs={"input": 0b0100})
        >>> sim.inspect("output")
        1

    :param vector: Takes a single arbitrary length :class:`WireVector`.

    :return: Returns a 1-bit result, the bitwise ``|`` of all of the bits in ``vector``.
    """
    return tree_reduce(lambda a, b: a | b, vector)


def xor_all_bits(vector: WireVector) -> WireVector:
    """Returns the result of bitwise XORing all the bits in ``vector``.

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> input = pyrtl.Input(name="input", bitwidth=4)
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.xor_all_bits(input)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input": 0b0100})
        >>> sim.inspect("output")
        1
        >>> sim.step(provided_inputs={"input": 0b0101})
        >>> sim.inspect("output")
        0

    :param vector: Takes a single arbitrary length :class:`WireVector`.

    :return: Returns a 1-bit result, the bitwise ``^`` of all of the bits in ``vector``.
    """
    return tree_reduce(lambda a, b: a ^ b, vector)


parity = xor_all_bits  # shadowing the xor_all_bits function
"""Alias for :func:`xor_all_bits`."""


def tree_reduce(op, vector: WireVector) -> WireVector:
    if len(vector) < 1:
        msg = "Cannot reduce empty vectors"
        raise PyrtlError(msg)
    if len(vector) == 1:
        return vector[0]
    left = tree_reduce(op, vector[: len(vector) // 2])
    right = tree_reduce(op, vector[len(vector) // 2 :])
    return op(left, right)


def _apply_op_over_all_bits(op, vector):
    if len(vector) < 1:
        msg = "Cannot reduce empty vectors"
        raise PyrtlError(msg)
    if len(vector) == 1:
        return vector[0]
    rest = _apply_op_over_all_bits(op, vector[1:])
    return op(vector[0], rest)


def rtl_any(*vectorlist: WireVectorLike) -> WireVector:
    """Hardware equivalent of Python's :func:`any`.

    Given any number of :class:`WireVectors<WireVector>`, return a 1-bit
    :class:`WireVector` which will hold a ``1`` if any of the inputs are ``1``. In other
    words, this generates a large OR gate. If no inputs are provided, it will return a
    :class:`Const` ``0`` (since there are no ``1s`` present) similar to Python's
    :func:`any` called with an empty list.

    .. note::

        ``rtl_any`` is most useful when working with a variable number of
        :class:`WireVectors<WireVector>`. For a fixed number of
        :class:`WireVectors<WireVector>`, it is clearer to use ``|``::

            any_ones = a | b | c

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> inputs = [pyrtl.Input(name="input0", bitwidth=1),
        ...           pyrtl.Input(name="input1", bitwidth=1)]
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtl_any(*inputs)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input0": 0, "input1": 0})
        >>> sim.inspect("output")
        0
        >>> sim.step(provided_inputs={"input0": 0, "input1": 1})
        >>> sim.inspect("output")
        1

    :param vectorlist: All arguments are length 1 :class:`WireVector`, or any type that
        can be coerced to :class:`WireVector` by :func:`as_wires`, with length 1.

    :raise PyrtlError: If any argument's :attr:`~WireVector.bitwidth` is not 1.

    :return: Length 1 :class:`WireVector` indicating if any bits in ``vectorlist`` are
             ``1``.
    """
    if len(vectorlist) == 0:
        return as_wires(False)
    converted_vectorlist = [as_wires(v) for v in vectorlist]
    if any(len(v) != 1 for v in converted_vectorlist):
        msg = "only length 1 WireVectors can be inputs to rtl_any"
        raise PyrtlError(msg)
    return or_all_bits(concat_list(converted_vectorlist))


def rtl_all(*vectorlist: WireVectorLike) -> WireVector:
    """Hardware equivalent of Python's :func:`all`.

    Given any number of :class:`WireVectors<WireVector>`, return a 1-bit
    :class:`WireVector` which will hold a ``1`` only if all of the inputs are ``1``. In
    other words, this generates a large AND gate. If no inputs are provided, it will
    return a :class:`Const` ``1`` (since there are no ``0s`` present) similar to
    Python's :func:`all` called with an empty list.

    .. note::

        ``rtl_all`` is most useful when working with a variable number of
        :class:`WireVectors<WireVector>`. For a fixed number of
        :class:`WireVectors<WireVector>`, it is clearer to use ``&``::

            all_ones = a & b & c

    .. doctest only::

        >>> import pyrtl
        >>> pyrtl.reset_working_block()

    Example::

        >>> inputs = [pyrtl.Input(name="input0", bitwidth=1),
        ...           pyrtl.Input(name="input1", bitwidth=1)]
        >>> output = pyrtl.Output(name="output")

        >>> output <<= pyrtl.rtl_all(*inputs)

        >>> sim = pyrtl.Simulation()
        >>> sim.step(provided_inputs={"input0": 0, "input1": 1})
        >>> sim.inspect("output")
        0
        >>> sim.step(provided_inputs={"input0": 1, "input1": 1})
        >>> sim.inspect("output")
        1

    :param vectorlist: All arguments are length 1 :class:`WireVector`, or any type that
        can be coerced to :class:`WireVector` by :func:`as_wires`, with length 1.

    :raise PyrtlError: If any argument's :attr:`~WireVector.bitwidth` is not 1.

    :return: Length 1 :class:`WireVector` indicating if all bits in ``vectorlist`` are
             ``1``.
    """
    if len(vectorlist) == 0:
        return as_wires(True)
    converted_vectorlist = [as_wires(v) for v in vectorlist]
    if any(len(v) != 1 for v in converted_vectorlist):
        msg = "only length 1 WireVectors can be inputs to rtl_all"
        raise PyrtlError(msg)
    return and_all_bits(concat_list(converted_vectorlist))


def _basic_mult(A, B):
    """A stripped-down copy of the Wallace multiplier in rtllib"""
    if len(B) == 1:
        A, B = B, A  # so that we can reuse the code below :)
    if len(A) == 1:
        return concat_list(
            [A & b for b in B] + [Const(0)]
        )  # keep WireVector len consistent

    result_bitwidth = len(A) + len(B)
    bits = [[] for weight in range(result_bitwidth)]
    for i, a in enumerate(A):
        for j, b in enumerate(B):
            bits[i + j].append(a & b)

    while not all(len(i) <= 2 for i in bits):
        deferred = [[] for weight in range(result_bitwidth + 1)]
        for i, w_array in enumerate(bits):  # Start with low weights and start reducing
            while len(w_array) >= 3:  # build a new full adder
                a, b, cin = (w_array.pop(0) for j in range(3))
                deferred[i].append(a ^ b ^ cin)
                deferred[i + 1].append(a & b | a & cin | b & cin)
            if len(w_array) == 2:
                a, b = w_array
                deferred[i].append(a ^ b)
                deferred[i + 1].append(a & b)
            else:
                deferred[i].extend(w_array)
        bits = deferred[:result_bitwidth]

    add_wires = tuple(itertools.zip_longest(*bits, fillvalue=Const(0)))
    adder_result = concat_list(add_wires[0]) + concat_list(add_wires[1])
    return adder_result[:result_bitwidth]


def _one_bit_add(a, b, carry_in):
    assert len(a) == len(b) == 1
    sumbit = a ^ b ^ carry_in
    carry_out = a & b | a & carry_in | b & carry_in
    return sumbit, carry_out


def _add_helper(a, b, carry_in):
    a, b = match_bitwidth(a, b)
    if len(a) == 1:
        sumbits, carry_out = _one_bit_add(a, b, carry_in)
    else:
        lsbit, ripplecarry = _one_bit_add(a[0], b[0], carry_in)
        msbits, carry_out = _add_helper(a[1:], b[1:], ripplecarry)
        sumbits = concat(msbits, lsbit)
    return sumbits, carry_out


def _basic_add(a, b):
    sumbits, carry_out = _add_helper(a, b, 0)
    return concat(carry_out, sumbits)


def _basic_sub(a, b):
    sumbits, carry_out = _add_helper(a, ~b, 1)
    return concat(carry_out, sumbits)


def _basic_eq(a, b):
    return ~or_all_bits(a ^ b)


def _basic_lt(a, b):
    assert len(a) == len(b)
    a_msb = a[-1]
    b_msb = b[-1]
    if len(a) == 1:
        return b_msb & ~a_msb
    small = _basic_lt(a[:-1], b[:-1])
    return (b_msb & ~a_msb) | (small & ~(a_msb ^ b_msb))


def _basic_gt(a, b):
    return _basic_lt(b, a)


def _basic_select(s, a, b):
    assert len(a) == len(b)
    assert len(s) == 1
    sa = concat(*[~s] * len(a))
    sb = concat(*[s] * len(b))
    return (a & sa) | (b & sb)
